from hiten.algorithms.hamiltonian.center._lie import \
    _lie_transform as _lie_transform_partial
from hiten.algorithms.hamiltonian.normal._lie import \
    _lie_transform as _lie_transform_full
from hiten.algorithms.hamiltonian.transforms import (
    _polylocal2realmodal, _polyrealmodal2local,
    _restrict_poly_to_center_manifold, _substitute_complex, _substitute_real)
from hiten.system.hamiltonians.base import (_CONVERSION_REGISTRY, Hamiltonian,
                                            LieGeneratingFunction)
from hiten.system.libration.collinear import CollinearPoint
from hiten.system.libration.triangular import TriangularPoint


def register_conversion(src_name: str, dst: "type[Hamiltonian] | str", 
                       required_context: list = None,
                       default_params: dict = None):
    """Decorator to register a *src* -> *dst* conversion.

    Parameters
    ----------
    src_name : str
        The ``Hamiltonian.name`` of the *source* representation.
    dst : type[Hamiltonian] | str
        Either the **class** for the destination representation *or* its
        ``name`` string.  Allowing a string lets you register conversions
        before defining dedicated subclasses.
    required_context : list, optional
        List of required context keys (e.g., ["point"] for LibrationPoint).
        These must be provided in the conversion call.
    default_params : dict, optional
        Default parameter values (e.g., {"tol": 1e-12}).

    Example
    -------
    >>> @register_conversion("physical", "real_modal", 
    ...                     required_context=["point"],
    ...                     default_params={"tol": 1e-12})
    ... def _physical_to_real(src: Hamiltonian, **kwargs) -> Hamiltonian:
    ...     point = kwargs["point"]
    ...     tol = kwargs.get("tol", 1e-12)
    ...     ...
    """

    dst_name: str
    if isinstance(dst, str):
        dst_name = dst
    else:
        dst_name = dst.name

    def _decorator(func):
        _CONVERSION_REGISTRY[(src_name, dst_name)] = (
            func, 
            required_context or [], 
            default_params or {}
        )
        return func

    return _decorator


@register_conversion("physical", "real_modal", 
                    required_context=["point"],
                    default_params={"tol": 1e-12})
def _physical_to_real_modal(ham: Hamiltonian, **kwargs) -> Hamiltonian:
    point = kwargs["point"]
    tol = kwargs.get("tol", 1e-12)
    new_poly = _polylocal2realmodal(point, ham.poly_H, ham.degree, ham._psi, ham._clmo, tol=tol)
    return Hamiltonian(new_poly, ham.degree, ham._ndof, name="real_modal")


@register_conversion("real_modal", "physical", 
                    required_context=["point"],
                    default_params={"tol": 1e-12})
def _real_modal_to_physical(ham: Hamiltonian, **kwargs) -> Hamiltonian:
    point = kwargs["point"]
    tol = kwargs.get("tol", 1e-12)
    new_poly = _polyrealmodal2local(point, ham.poly_H, ham.degree, ham._psi, ham._clmo, tol=tol)
    return Hamiltonian(new_poly, ham.degree, ham._ndof, name="physical")


@register_conversion("real_modal", "complex_modal", 
                    required_context=["point"],
                    default_params={"tol": 1e-12})
def _real_modal_to_complex_modal(ham: Hamiltonian, **kwargs) -> Hamiltonian:
    point = kwargs["point"]
    if isinstance(point, CollinearPoint):
        mix_pairs = (1, 2)
    elif isinstance(point, TriangularPoint):
        mix_pairs = (0, 1, 2)

    tol = kwargs.get("tol", 1e-12)
    new_poly = _substitute_complex(ham.poly_H, ham.degree, ham._psi, ham._clmo, tol=tol, mix_pairs=mix_pairs)
    return Hamiltonian(new_poly, ham.degree, ham._ndof, name="complex_modal")


@register_conversion("complex_modal", "real_modal", 
                    required_context=["point"],
                    default_params={"tol": 1e-12})
def _complex_modal_to_real_modal(ham: Hamiltonian, **kwargs) -> Hamiltonian:
    point = kwargs["point"]
    if isinstance(point, CollinearPoint):
        mix_pairs = (1, 2)
    elif isinstance(point, TriangularPoint):
        mix_pairs = (0, 1, 2)

    tol = kwargs.get("tol", 1e-12)
    new_poly = _substitute_real(ham.poly_H, ham.degree, ham._psi, ham._clmo, tol=tol, mix_pairs=mix_pairs)
    return Hamiltonian(new_poly, ham.degree, ham._ndof, name="real_modal")


@register_conversion("complex_modal", "complex_partial_normal", 
                    required_context=["point"],
                    default_params={"tol_lie": 1e-30})
def _complex_modal_to_complex_partial_normal(ham: Hamiltonian, **kwargs) -> tuple[Hamiltonian, "LieGeneratingFunction"]:
    point = kwargs["point"]
    tol_lie = kwargs.get("tol_lie", 1e-30)
    # This returns (poly_trans, poly_G_total, poly_elim_total)
    new_poly, poly_G_total, poly_elim_total = _lie_transform_partial(point, ham.poly_H, ham._psi, ham._clmo, ham.degree, tol=tol_lie)
    
    new_ham = Hamiltonian(new_poly, ham.degree, ham._ndof, name="complex_partial_normal")
    generating_functions = LieGeneratingFunction(poly_G=poly_G_total, poly_elim=poly_elim_total, degree=ham.degree, ndof=ham._ndof, name="generating_functions_partial")
    
    return new_ham, generating_functions


@register_conversion("complex_partial_normal", "real_partial_normal", 
                    required_context=["point"],
                    default_params={"tol": 1e-14})
def _complex_partial_normal_to_real_partial_normal(ham: Hamiltonian, **kwargs) -> Hamiltonian:
    point = kwargs["point"]
    if isinstance(point, CollinearPoint):
        mix_pairs = (1, 2)
    elif isinstance(point, TriangularPoint):
        mix_pairs = (0, 1, 2)

    tol = kwargs.get("tol", 1e-14)
    new_poly = _substitute_real(ham.poly_H, ham.degree, ham._psi, ham._clmo, tol=tol, mix_pairs=mix_pairs)
    
    return Hamiltonian(new_poly, ham.degree, ham._ndof, name="real_partial_normal")


@register_conversion("real_partial_normal", "complex_partial_normal", 
                    required_context=["point"],
                    default_params={"tol": 1e-14})
def _real_partial_normal_to_complex_partial_normal(ham: Hamiltonian, **kwargs) -> Hamiltonian:
    point = kwargs["point"]
    if isinstance(point, CollinearPoint):
        mix_pairs = (1, 2)
    elif isinstance(point, TriangularPoint):
        mix_pairs = (0, 1, 2)

    tol = kwargs.get("tol", 1e-14)
    new_poly = _substitute_complex(ham.poly_H, ham.degree, ham._psi, ham._clmo, tol=tol, mix_pairs=mix_pairs)
    return Hamiltonian(new_poly, ham.degree, ham._ndof, name="complex_partial_normal")


@register_conversion("complex_partial_normal", "center_manifold_complex", 
                    required_context=["point"],
                    default_params={"tol": 1e-14})
def _complex_partial_normal_to_center_manifold_complex(ham: Hamiltonian, **kwargs) -> Hamiltonian:
    point = kwargs["point"]
    tol = kwargs.get("tol", 1e-14)
    new_poly = _restrict_poly_to_center_manifold(point, ham.poly_H, ham._clmo, tol=tol)
    return Hamiltonian(new_poly, ham.degree, ham._ndof, name="center_manifold_complex")


@register_conversion("center_manifold_complex", "center_manifold_real", 
                    required_context=["point"],
                    default_params={"tol": 1e-14})
def _center_manifold_complex_to_center_manifold_real(ham: Hamiltonian, **kwargs) -> Hamiltonian:
    point = kwargs["point"]
    if isinstance(point, CollinearPoint):
        mix_pairs = (1, 2)
    elif isinstance(point, TriangularPoint):
        mix_pairs = (0, 1, 2)

    tol = kwargs.get("tol", 1e-14)
    new_poly = _substitute_real(ham.poly_H, ham.degree, ham._psi, ham._clmo, tol=tol, mix_pairs=mix_pairs)
    return Hamiltonian(new_poly, ham.degree, ham._ndof, name="center_manifold_real")


@register_conversion("center_manifold_real", "center_manifold_complex", 
                    required_context=["point"],
                    default_params={"tol": 1e-14})
def _center_manifold_real_to_center_manifold_complex(ham: Hamiltonian, **kwargs) -> Hamiltonian:
    point = kwargs["point"]
    if isinstance(point, CollinearPoint):
        mix_pairs = (1, 2)
    elif isinstance(point, TriangularPoint):
        mix_pairs = (0, 1, 2)

    tol = kwargs.get("tol", 1e-14)
    new_poly = _substitute_complex(ham.poly_H, ham.degree, ham._psi, ham._clmo, tol=tol, mix_pairs=mix_pairs)
    return Hamiltonian(new_poly, ham.degree, ham._ndof, name="center_manifold_complex")


@register_conversion("complex_modal", "complex_full_normal", 
                    required_context=["point"],
                    default_params={"tol_lie": 1e-30, "resonance_tol": 1e-14})
def _complex_modal_to_complex_full_normal(ham: Hamiltonian, **kwargs) -> tuple[Hamiltonian, "LieGeneratingFunction"]:
    point = kwargs["point"]
    tol_lie = kwargs.get("tol_lie", 1e-30)
    resonance_tol = kwargs.get("resonance_tol", 1e-14)
    new_poly, poly_G_total, poly_elim_total = _lie_transform_full(point, ham.poly_H, ham._psi, ham._clmo, ham.degree, tol=tol_lie, resonance_tol=resonance_tol)
    
    new_ham = Hamiltonian(new_poly, ham.degree, ham._ndof, name="complex_full_normal")
    generating_functions = LieGeneratingFunction(poly_G=poly_G_total, poly_elim=poly_elim_total, degree=ham.degree, ndof=ham._ndof, name="generating_functions_full")
    
    return new_ham, generating_functions


@register_conversion("complex_full_normal", "real_full_normal", 
                    required_context=["point"],
                    default_params={"tol": 1e-14})
def _complex_full_normal_to_real_full_normal(ham: Hamiltonian, **kwargs) -> Hamiltonian:
    point = kwargs["point"]
    if isinstance(point, CollinearPoint):
        mix_pairs = (1, 2)
    elif isinstance(point, TriangularPoint):
        mix_pairs = (0, 1, 2)

    tol = kwargs.get("tol", 1e-14)
    new_poly = _substitute_real(ham.poly_H, ham.degree, ham._psi, ham._clmo, tol=tol, mix_pairs=mix_pairs)
    return Hamiltonian(new_poly, ham.degree, ham._ndof, name="real_full_normal")


@register_conversion("real_full_normal", "complex_full_normal", 
                    required_context=["point"],
                    default_params={"tol": 1e-14})
def _real_full_normal_to_complex_full_normal(ham: Hamiltonian, **kwargs) -> Hamiltonian:
    point = kwargs["point"]
    if isinstance(point, CollinearPoint):
        mix_pairs = (1, 2)
    elif isinstance(point, TriangularPoint):
        mix_pairs = (0, 1, 2)

    tol = kwargs.get("tol", 1e-14)
    new_poly = _substitute_complex(ham.poly_H, ham.degree, ham._psi, ham._clmo, tol=tol, mix_pairs=mix_pairs)
    return Hamiltonian(new_poly, ham.degree, ham._ndof, name="complex_full_normal")


import random

import numpy as np
import pytest
from numba import types
from numba.typed import Dict, List

from hiten.algorithms.polynomial.base import (_create_encode_dict_from_clmo,
                                               _decode_multiindex,
                                               _encode_multiindex,
                                               _init_index_tables, _make_poly)
from hiten.algorithms.utils.config import N_VARS

TEST_MAX_DEG = 5
PSI, CLMO = _init_index_tables(TEST_MAX_DEG)
ENCODE_DICT = _create_encode_dict_from_clmo(CLMO)

def test_init_index_tables():
    """Test if the index tables are initialized correctly"""
    # Check dimensions of psi
    assert PSI.shape == (N_VARS+1, TEST_MAX_DEG+1)
    
    # Check a few known values of psi (binomial coefficients)
    assert PSI[1, 3] == 1  # Variables=1, Degree=3: only x^3
    assert PSI[2, 2] == 3  # Variables=2, Degree=2: x^2, xy, y^2
    assert PSI[3, 1] == 3  # Variables=3, Degree=1: x, y, z
    
    # Check clmo list length
    assert len(CLMO) == TEST_MAX_DEG + 1
    
    # Check sizes of clmo arrays for different degrees
    for d in range(TEST_MAX_DEG + 1):
        assert len(CLMO[d]) == PSI[N_VARS, d]

def test_make_poly():
    """Test creation of zero polynomials with complex128 dtype"""
    for degree in range(TEST_MAX_DEG + 1):
        # Test polynomial creation
        poly = _make_poly(degree, PSI)
        
        # Check size
        expected_size = PSI[N_VARS, degree]
        assert poly.shape[0] == expected_size
        
        # Check if all coefficients are zero
        assert np.all(poly == 0.0)
        
        # Check data type is always complex128
        assert poly.dtype == np.complex128

def test_decode_multiindex():
    """Test decoding multiindices"""
    for degree in range(1, TEST_MAX_DEG + 1):
        # Get size of the polynomial for this degree
        size = PSI[N_VARS, degree]
        
        if size <= 50:
            positions = range(size)
        else:
            positions = list(range(20))
            positions.extend(range(size-20, size))
            positions.extend(range(size//2-5, size//2+5))
        
        for pos in positions:
            k = _decode_multiindex(pos, degree, CLMO)
            
            k = np.asarray(k)
            assert k.shape == (N_VARS,)
            
            assert np.sum(k) == degree
            
            assert np.all(k >= 0)
            
            assert np.all(k <= degree)

def test_encode_multiindex():
    """Test encoding multiindices"""
    for degree in range(1, TEST_MAX_DEG + 1):
        size = PSI[N_VARS, degree]
        
        if size <= 50:
            positions = range(size)
        else:
            positions = list(range(20))
            positions.extend(range(size-20, size))
            positions.extend(range(size//2-5, size//2+5))
        
        for pos in positions:
            k = _decode_multiindex(pos, degree, CLMO)
            k = np.asarray(k)
            
            idx = _encode_multiindex(k, degree, ENCODE_DICT)
            assert idx == pos
            
            if np.any(k > 0):
                k_invalid = k.copy()
                for i in range(N_VARS):
                    if k[i] > 0:
                        k_invalid[i] -= 1
                        k_invalid[(i+1) % N_VARS] += 1
                        idx_invalid = _encode_multiindex(k_invalid, degree, ENCODE_DICT)
                        assert idx_invalid != pos
                        break

def test_multi_index_roundtrip():
    """Test full encode-decode roundtrip for multiindices"""
    for degree in range(TEST_MAX_DEG + 1):
        size = PSI[N_VARS, degree]
        
        if degree <= 2:
            test_positions = range(size)
        else:
            test_positions = []
            step = max(1, size // 20)
            for i in range(0, size, step):
                test_positions.append(i)
            if size > 0:
                test_positions.append(0)
                test_positions.append(size - 1)
                
        for pos in test_positions:
            k1 = _decode_multiindex(pos, degree, CLMO)
            idx = _encode_multiindex(k1, degree, ENCODE_DICT)
            k2 = _decode_multiindex(idx, degree, CLMO)
            
            assert idx == pos
            
            np.testing.assert_array_equal(k1, k2)


@pytest.mark.parametrize("deg", [0, 1, 2, 3, 4, 5, 6])
def test_encode_decode_roundtrip(deg):
    _, clmo = _init_index_tables(6)
    encode_dict_for_deg6 = _create_encode_dict_from_clmo(clmo)
    for _ in range(10):
        k = np.zeros(N_VARS, dtype=np.int64)
        remaining = deg
        for i in range(N_VARS - 1):
            k[i] = random.randint(0, remaining)
            remaining -= k[i]
        k[-1] = remaining
        idx = _encode_multiindex(k, deg, encode_dict_for_deg6)
        assert idx >= 0
        k_back = _decode_multiindex(idx, deg, clmo)
        assert np.array_equal(k_back, k)

    k = np.array([0,0,0,0,0,0], dtype=np.int64)
    deg = 0
    _, clmo = _init_index_tables(deg)
    local_encode_dict_list = List.empty_list(types.DictType(types.int64, types.int32))
    for d_arr in clmo:
        d_map = Dict.empty(key_type=types.int64, value_type=types.int32)
        for i, p_val in enumerate(d_arr):
            d_map[np.int64(p_val)] = np.int32(i)
        local_encode_dict_list.append(d_map)

    idx = _encode_multiindex(k, deg, local_encode_dict_list)
    assert idx == 0



import json
import pickle
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict

import h5py
import numpy as np

from hiten.utils.io.common import _ensure_dir
from hiten.utils.io.map import save_poincare_map, load_poincare_map

if TYPE_CHECKING:
    from hiten.system.center import CenterManifold
    from hiten.algorithms.poincare.centermanifold.base import CenterManifoldMap
    from hiten.system.hamiltonians.pipeline import HamiltonianPipeline

__all__ = ["save_center_manifold", "load_center_manifold"]

HDF5_VERSION = "2.0"

def _serialize_hamiltonians(pipeline: "HamiltonianPipeline") -> bytes:
    """Return a pickled representation of the pipeline's Hamiltonian cache."""
    data: Dict[str, Dict[str, Any]] = {}
    for form, ham in pipeline._hamiltonian_cache.items():
        data[form] = {
            "poly_H": ham.poly_H,
            "degree": ham.degree,
            "ndof": ham.ndof,
            "name": ham.name,
        }
    return pickle.dumps(data, protocol=pickle.HIGHEST_PROTOCOL)


def _deserialize_hamiltonians(blob: bytes, pipeline: "HamiltonianPipeline") -> None:
    from hiten.system.hamiltonians.base import Hamiltonian

    cache_data: Dict[str, Dict[str, Any]] = pickle.loads(blob)
    for form, info in cache_data.items():
        ham = Hamiltonian(info["poly_H"], info["degree"], ndof=info["ndof"], name=info["name"])
        pipeline._hamiltonian_cache[form] = ham


def save_center_manifold(
    cm: "CenterManifold",
    dir_path: str | Path,
    *,
    include_hamiltonians: bool = False,
    compression: str = "gzip",
    level: int = 4,
) -> None:
    """Serialise *cm* and (optionally) cached Hamiltonians.

    Parameters
    ----------
    include_hamiltonians
        If *True* the cached Hamiltonian objects stored in the internal
        pipeline are pickled and saved alongside the core data.  This speeds
        up subsequent loads at the cost of larger files.
    """
    from hiten.utils.io.common import _write_dataset

    dir_path = Path(dir_path)
    _ensure_dir(dir_path)

    main_file = dir_path / "manifold.h5"
    with h5py.File(main_file, "w") as f:
        f.attrs["format_version"] = HDF5_VERSION
        f.attrs["class"] = cm.__class__.__name__
        f.attrs["degree"] = int(cm.degree)

        # Serialize point via pickle (small)
        point_blob = pickle.dumps(cm.point, protocol=pickle.HIGHEST_PROTOCOL)
        f.create_dataset("point_pickle", data=np.frombuffer(point_blob, dtype=np.uint8))

        if include_hamiltonians and cm.pipeline._hamiltonian_cache:
            ham_blob = _serialize_hamiltonians(cm.pipeline)
            _write_dataset(f, "hamiltonians_pickle", np.frombuffer(ham_blob, dtype=np.uint8))

    if cm._poincare_maps:
        maps_dir = dir_path / "maps"
        maps_dir.mkdir(exist_ok=True)
        keys_info = []
        for idx, (key, pmap) in enumerate(cm._poincare_maps.items()):
            filename = f"map_{idx}.h5"
            save_poincare_map(pmap, maps_dir / filename, compression=compression, level=level)
            keys_info.append({"key": list(key), "file": filename})
        (dir_path / "poincare_maps_keys.json").write_text(json.dumps(keys_info))
    else:
        # Remove stale files if present
        key_file = dir_path / "poincare_maps_keys.json"
        if key_file.exists():
            key_file.unlink()
        maps_dir = dir_path / "maps"
        if maps_dir.exists() and not any(maps_dir.iterdir()):
            maps_dir.rmdir()


def load_center_manifold(dir_path: str | Path):
    """Return a reconstructed :class:`CenterManifold`."""
    from hiten.system.center import CenterManifold

    dir_path = Path(dir_path)
    main_file = dir_path / "manifold.h5"
    if not main_file.exists():
        raise FileNotFoundError(main_file)

    with h5py.File(main_file, "r") as f:
        degree = int(f.attrs["degree"])
        point_blob = f["point_pickle"][()]
        point = pickle.loads(point_blob.tobytes())

        cm = CenterManifold(point, degree)

        if "hamiltonians_pickle" in f:
            ham_blob = f["hamiltonians_pickle"][()]
            _deserialize_hamiltonians(ham_blob.tobytes(), cm.pipeline)
            # Refresh internal Hamiltonian system reference to the deserialized one
            cm._hamsys = cm.pipeline.get_hamiltonian("center_manifold_real").hamsys

    maps_key_file = dir_path / "poincare_maps_keys.json"
    maps_dir = dir_path / "maps"
    if maps_key_file.exists():
        keys_info = json.loads(maps_key_file.read_text())
        for info in keys_info:
            key_list = info["key"]
            energy = key_list[0]
            config_tuple = tuple(tuple(item) for item in key_list[1])
            cache_key = (energy, config_tuple)

            pmap = load_poincare_map(maps_dir / info["file"], cm)
            cm._poincare_maps[cache_key] = pmap

    return cm



import numpy as np
import pytest

from hiten.algorithms.polynomial.algebra import (_get_degree, _poly_add,
                                                  _poly_clean,
                                                  _poly_clean_inplace,
                                                  _poly_diff, _poly_evaluate,
                                                  _poly_mul, _poly_poisson,
                                                  _poly_scale)
from hiten.algorithms.polynomial.base import (_ENCODE_DICT_GLOBAL,
                                               _decode_multiindex,
                                               _encode_multiindex,
                                               _init_index_tables, _make_poly)
from hiten.algorithms.utils.config import N_VARS

TEST_MAX_DEG = 5
PSI, CLMO = _init_index_tables(TEST_MAX_DEG)


def test_poly_add():
    for degree in range(TEST_MAX_DEG + 1):
        a = _make_poly(degree, PSI)
        b = _make_poly(degree, PSI)
        result = _make_poly(degree, PSI)
        
        if degree == 0:
            a[0] = 1.5
            b[0] = 2.5
            _poly_add(a, b, result)
            assert result[0] == 4.0
            continue
        
        size = PSI[N_VARS, degree]
        
        for i in range(min(size, 10)):
            a[i] = i * 1.5
            b[i] = i * 0.5
        
        zero_poly = _make_poly(degree, PSI)
        result_zero = _make_poly(degree, PSI)
        _poly_add(a, zero_poly, result_zero)
        np.testing.assert_array_equal(a, result_zero)
        
        expected = np.zeros_like(a)
        for i in range(a.shape[0]):
            expected[i] = a[i] + b[i]
        
        _poly_add(a, b, result)
        np.testing.assert_array_almost_equal(result, expected)
        
        result_commute = _make_poly(degree, PSI)
        _poly_add(b, a, result_commute)
        np.testing.assert_array_almost_equal(result, result_commute)
        
        double_a = _make_poly(degree, PSI)
        _poly_add(a, a, double_a)
        
        expected_double = _make_poly(degree, PSI)
        _poly_scale(a, 2.0, expected_double)
        np.testing.assert_array_almost_equal(double_a, expected_double)

def test_add_negative_numbers():
    degree = 3
    size = PSI[N_VARS, degree]
    
    a = _make_poly(degree, PSI)
    b = _make_poly(degree, PSI)
    result = _make_poly(degree, PSI)
    
    # Set alternating positive and negative coefficients
    for i in range(size):
        a[i] = (-1)**i * (i + 1.0)  # 1, -2, 3, -4, ...
        b[i] = (-1)**(i+1) * (i + 2.0)  # -2, 3, -4, 5, ...
    
    # Expected: coefficients should cancel out to (-1, 1, -1, 1, ...)
    expected = np.zeros_like(a)
    for i in range(size):
        expected[i] = a[i] + b[i]
    
    _poly_add(a, b, result)
    np.testing.assert_array_almost_equal(result, expected)

def test_add_large_values():
    """Test polynomial addition with very large coefficients"""
    degree = 3
    size = PSI[N_VARS, degree]
    
    a = _make_poly(degree, PSI)
    b = _make_poly(degree, PSI)
    result = _make_poly(degree, PSI)
    
    # Set large coefficients
    for i in range(size):
        a[i] = 1e15 + i
        b[i] = 2e15 - i
    
    # Expected result
    expected = np.zeros_like(a)
    for i in range(size):
        expected[i] = a[i] + b[i]
    
    _poly_add(a, b, result)
    np.testing.assert_array_almost_equal(result, expected)

def test_add_small_values():
    """Test polynomial addition with very small coefficients"""
    degree = 3
    size = PSI[N_VARS, degree]
    
    a = _make_poly(degree, PSI)
    b = _make_poly(degree, PSI)
    result = _make_poly(degree, PSI)
    
    # Set small coefficients
    for i in range(size):
        a[i] = 1e-15 * (i + 1)
        b[i] = 2e-15 * (i + 1)
    
    # Expected result
    expected = np.zeros_like(a)
    for i in range(size):
        expected[i] = a[i] + b[i]
    
    _poly_add(a, b, result)
    np.testing.assert_array_almost_equal(result, expected)

def test_add_mixed_values():
    """Test polynomial addition with mixed signs and magnitudes"""
    degree = 3
    size = PSI[N_VARS, degree]
    
    a = _make_poly(degree, PSI)
    b = _make_poly(degree, PSI)
    result = _make_poly(degree, PSI)
    
    # Set coefficients with varying magnitudes and signs
    for i in range(size):
        if i % 3 == 0:
            a[i] = 1e10 * (i + 1)
            b[i] = -1e10 * i
        elif i % 3 == 1:
            a[i] = -1e-10 * (i + 1)
            b[i] = 1e-10 * i
        else:
            a[i] = (-1)**i * i
            b[i] = (-1)**(i+1) * (i + 1)
    
    # Expected result
    expected = np.zeros_like(a)
    for i in range(size):
        expected[i] = a[i] + b[i]
    
    _poly_add(a, b, result)
    np.testing.assert_array_almost_equal(result, expected)

def test_add_special_values():
    """Test polynomial addition with special values (inf, nan)"""
    degree = 3
    
    a = _make_poly(degree, PSI)
    b = _make_poly(degree, PSI)
    result = _make_poly(degree, PSI)
    
    # Set some special values
    a[0] = np.inf
    a[1] = -np.inf
    a[2] = np.nan
    
    b[0] = 1.0
    b[1] = -np.inf
    b[2] = 1.0
    
    _poly_add(a, b, result)
    
    # Check expected behavior for inf + finite = inf
    assert np.isinf(result[0])
    assert result[0] > 0  # positive infinity
    
    # Check expected behavior for -inf + (-inf) = -inf
    assert np.isinf(result[1])
    assert result[1] < 0  # negative infinity
    
    # Check expected behavior for nan + anything = nan
    assert np.isnan(result[2])

def test_add_complex_numbers():
    """Test polynomial addition with complex coefficients"""
    degree = 3
    size = PSI[N_VARS, degree]
    
    # Create complex polynomials
    a = _make_poly(degree, PSI)
    b = _make_poly(degree, PSI)
    result = _make_poly(degree, PSI)
    
    # Set complex coefficients in a
    for i in range(size):
        a[i] = complex(i, i+1)  # a[i] = i + (i+1)j
    
    # Set complex coefficients in b
    for i in range(size):
        b[i] = complex(2*i, -(i+1))  # b[i] = 2i - (i+1)j
    
    # Expected result: a[i] + b[i] = (3i) + 0j
    expected = np.zeros_like(a)
    for i in range(size):
        expected[i] = a[i] + b[i]
    
    # Perform addition
    _poly_add(a, b, result)
    
    # Verify result
    np.testing.assert_array_almost_equal(result, expected)
    
    # Verify specific values
    for i in range(size):
        # Real part should be i + 2i = 3i
        assert result[i].real == 3*i
        # Imaginary part should be (i+1) - (i+1) = 0
        assert abs(result[i].imag) < 1e-10
    
    # Test with different patterns of complex numbers
    a = _make_poly(degree, PSI)
    b = _make_poly(degree, PSI)
    result = _make_poly(degree, PSI)
    
    # Pattern 1: Alternating real and imaginary
    for i in range(size):
        if i % 2 == 0:
            a[i] = complex(i, 0)  # pure real
            b[i] = complex(0, i)  # pure imaginary
        else:
            a[i] = complex(0, i)  # pure imaginary
            b[i] = complex(i, 0)  # pure real
    
    # Expected: all coefficients should have both real and imaginary parts = i
    expected = np.zeros_like(a)
    for i in range(size):
        expected[i] = a[i] + b[i]
    
    _poly_add(a, b, result)
    np.testing.assert_array_almost_equal(result, expected)
    
    # Test with complex conjugates
    a = _make_poly(degree, PSI)
    b = _make_poly(degree, PSI)
    result = _make_poly(degree, PSI)
    
    # a and b are complex conjugates of each other
    for i in range(size):
        a[i] = complex(i, i+1)
        b[i] = complex(i, -(i+1))  # complex conjugate of a[i]
    
    # Expected: result should be purely real
    expected = np.zeros_like(a)
    for i in range(size):
        expected[i] = a[i] + b[i]
    
    _poly_add(a, b, result)
    np.testing.assert_array_almost_equal(result, expected)
    
    # Verify all results have zero imaginary part
    for i in range(size):
        assert abs(result[i].imag) < 1e-10
        assert result[i].real == 2*i  # real part should be 2*i

def test_poly_scale_basic():
    """Test basic polynomial scaling with various factors"""
    degree = 3
    size = PSI[N_VARS, degree]
    
    # Create test polynomial
    p = _make_poly(degree, PSI)
    
    # Set some coefficients
    for i in range(size):
        p[i] = i * 1.5
    
    # Test scaling by different factors
    factors = [0.0, 1.0, -1.0, 2.0, -0.5, 10.0, 0.1]
    
    for factor in factors:
        # Create result polynomial
        result = _make_poly(degree, PSI)
        
        # Scale the polynomial
        _poly_scale(p, factor, result)
        
        # Check results
        for i in range(size):
            expected = p[i] * factor
            assert result[i] == expected
            
        # Verify that original polynomial hasn't changed
        for i in range(size):
            assert p[i] == i * 1.5

def test_poly_scale_zero():
    """Test scaling polynomial by zero"""
    degree = 3
    
    # Create test polynomial with non-zero values
    p = _make_poly(degree, PSI)
    result = _make_poly(degree, PSI)
    
    # Set non-zero values
    for i in range(p.shape[0]):
        p[i] = 10.0 * (i + 1)
    
    # Scale by zero
    _poly_scale(p, 0.0, result)
    
    # All values should be zero
    assert np.all(result == 0.0)
    
    # Original should be unchanged
    for i in range(p.shape[0]):
        assert p[i] == 10.0 * (i + 1)

def test_poly_scale_large_values():
    """Test scaling with large factors and values"""
    degree = 3
    size = PSI[N_VARS, degree]
    
    # Create test polynomial
    p = _make_poly(degree, PSI)
    result = _make_poly(degree, PSI)
    
    # Test with large coefficients
    for i in range(size):
        p[i] = 1e8 * (i + 1)
    
    # Scale by large factor
    large_factor = 1e8
    _poly_scale(p, large_factor, result)
    
    # Check results
    for i in range(size):
        expected = p[i] * large_factor
        assert result[i] == expected

def test_poly_scale_small_values():
    """Test scaling with small factors and values"""
    degree = 3
    size = PSI[N_VARS, degree]
    
    # Create test polynomial
    p = _make_poly(degree, PSI)
    result = _make_poly(degree, PSI)
    
    # Test with small coefficients
    for i in range(size):
        p[i] = 1e-8 * (i + 1)
    
    # Scale by small factor
    small_factor = 1e-8
    _poly_scale(p, small_factor, result)
    
    # Check results
    expected = np.zeros_like(p)
    for i in range(size):
        expected[i] = p[i] * small_factor
    
    np.testing.assert_array_almost_equal(result, expected)

def test_poly_scale_special_values():
    """Test scaling with special values (inf, nan)"""
    degree = 3
    
    # Create test polynomial
    p = _make_poly(degree, PSI)
    result = _make_poly(degree, PSI)
    
    # Set special values
    p[0] = np.inf
    p[1] = -np.inf
    p[2] = np.nan
    p[3] = 1.0
    
    # Scale by positive value
    _poly_scale(p, 2.0, result)
    
    # Check results
    assert np.isinf(result[0])
    assert result[0] > 0  # positive infinity
    assert np.isinf(result[1])
    assert result[1] < 0  # negative infinity
    assert np.isnan(result[2])
    assert result[3] == 2.0
    
    # Scale by negative value
    _poly_scale(p, -1.0, result)
    
    # Check results
    assert np.isinf(result[0])
    assert result[0] < 0  # negative infinity (sign flipped)
    assert np.isinf(result[1])
    assert result[1] > 0  # positive infinity (sign flipped)
    assert np.isnan(result[2])
    assert result[3] == -1.0
    
    # Scale by zero
    _poly_scale(p, 0.0, result)
    
    # Check results (0 * inf = nan in IEEE 754)
    assert np.isnan(result[0])
    assert np.isnan(result[1])
    assert np.isnan(result[2])
    assert result[3] == 0.0

def test_poly_scale_complex():
    """Test scaling complex polynomials"""
    degree = 3
    size = PSI[N_VARS, degree]
    
    # Create complex polynomial
    p = _make_poly(degree, PSI)
    result = _make_poly(degree, PSI)
    
    # Set complex coefficients
    for i in range(size):
        p[i] = complex(i, i+1)
    
    # Test with real scalar
    real_factor = 2.0
    _poly_scale(p, real_factor, result)
    
    # Check results
    for i in range(size):
        expected = p[i] * real_factor
        assert result[i].real == expected.real
        assert result[i].imag == expected.imag
    
    # Test with complex scalar
    complex_factor = complex(1.0, 2.0)
    _poly_scale(p, complex_factor, result)
    
    # Check results
    for i in range(size):
        expected = p[i] * complex_factor
        assert abs(result[i].real - expected.real) < 1e-10
        assert abs(result[i].imag - expected.imag) < 1e-10
    
    # Test with zero complex scalar
    zero_complex = complex(0.0, 0.0)
    _poly_scale(p, zero_complex, result)
    
    # All values should be zero
    for i in range(size):
        assert result[i].real == 0.0
        assert result[i].imag == 0.0

def test_poly_mul_monomials():
    """Test multiplication of monomials (single-term polynomials)"""
    # Test with low degrees to keep the test easy to reason about
    deg_p = 1  # Linear monomial
    deg_q = 2  # Quadratic monomial
    
    p = _make_poly(deg_p, PSI)
    q = _make_poly(deg_q, PSI)
    
    # Set p to represent x0
    # Find the index for the monomial x0
    idx_x0 = _encode_multiindex(np.array([1, 0, 0, 0, 0, 0], dtype=np.int64), deg_p, _ENCODE_DICT_GLOBAL)
    p[idx_x0] = 2.0  # p = 2*x0
    
    # Set q to represent x0*x1
    # Find the index for the monomial x0*x1
    idx_x0x1 = _encode_multiindex(np.array([1, 1, 0, 0, 0, 0], dtype=np.int64), deg_q, _ENCODE_DICT_GLOBAL)
    q[idx_x0x1] = 3.0  # q = 3*x0*x1
    
    # Multiply the polynomials
    result = _poly_mul(p, deg_p, q, deg_q, PSI, CLMO, _ENCODE_DICT_GLOBAL) # Pass _ENCODE_DICT_GLOBAL
    
    # Verify the degree of the result
    assert result.shape[0] == PSI[N_VARS, deg_p + deg_q]
    
    # The result should be 6*x0^2*x1
    idx_x0x0x1 = _encode_multiindex(np.array([2, 1, 0, 0, 0, 0], dtype=np.int64), deg_p + deg_q, _ENCODE_DICT_GLOBAL)
    assert abs(result[idx_x0x0x1] - 6.0) < 1e-10
    
    # Verify all other terms are zero
    result[idx_x0x0x1] = 0.0
    assert np.all(result == 0.0)
    
    # Test with a different pair of monomials
    p = _make_poly(deg_p, PSI)
    q = _make_poly(deg_q, PSI)
    
    # p = 4*x1
    idx_x1 = _encode_multiindex(np.array([0, 1, 0, 0, 0, 0], dtype=np.int64), deg_p, _ENCODE_DICT_GLOBAL)
    p[idx_x1] = 4.0
    
    # q = 5*x2^2
    idx_x2sq = _encode_multiindex(np.array([0, 0, 2, 0, 0, 0], dtype=np.int64), deg_q, _ENCODE_DICT_GLOBAL)
    q[idx_x2sq] = 5.0
    
    # Multiply
    result = _poly_mul(p, deg_p, q, deg_q, PSI, CLMO, _ENCODE_DICT_GLOBAL) # Pass _ENCODE_DICT_GLOBAL
    
    # Result should be 20*x1*x2^2
    idx_result = _encode_multiindex(np.array([0, 1, 2, 0, 0, 0], dtype=np.int64), deg_p + deg_q, _ENCODE_DICT_GLOBAL)
    assert abs(result[idx_result] - 20.0) < 1e-10
    
    # Verify all other terms are zero
    result[idx_result] = 0.0
    assert np.all(result == 0.0)

def test_poly_mul_basic():
    """Test basic polynomial multiplication"""
    # Use small degrees to keep the test manageable
    deg_p = 1
    deg_q = 1
    
    p = _make_poly(deg_p, PSI) # PSI is the global from base.py via this file's init
    q = _make_poly(deg_q, PSI)
    
    # Set p = 2*x0 + 3*x1 (Homogeneous degree 1)
    # _encode_multiindex calls should use _ENCODE_DICT_GLOBAL as PSI, CLMO are the global ones here.
    idx_x0 = _encode_multiindex(np.array([1, 0, 0, 0, 0, 0], dtype=np.int64), deg_p, _ENCODE_DICT_GLOBAL)
    idx_x1 = _encode_multiindex(np.array([0, 1, 0, 0, 0, 0], dtype=np.int64), deg_p, _ENCODE_DICT_GLOBAL)
    
    p[idx_x0] = 2.0
    p[idx_x1] = 3.0
    
    # Set q = 5*x0 + 6*x1 (Homogeneous degree 1)
    # idx_x0q should be same as idx_x0, idx_x1q same as idx_x1 as degrees are same.
    q[idx_x0] = 5.0 
    q[idx_x1] = 6.0
    
    # Multiply the polynomials
    result = _poly_mul(p, deg_p, q, deg_q, PSI, CLMO, _ENCODE_DICT_GLOBAL) # Pass _ENCODE_DICT_GLOBAL
    
    # Expected terms in the result for (2x0+3x1)*(5x0+6x1):
    # 2*x0 * 5*x0 = 10*x0^2
    # 2*x0 * 6*x1 = 12*x0*x1
    # 3*x1 * 5*x0 = 15*x0*x1
    # 3*x1 * 6*x1 = 18*x1^2
    
    idx_x0sq = _encode_multiindex(np.array([2, 0, 0, 0, 0, 0], dtype=np.int64), deg_p + deg_q, _ENCODE_DICT_GLOBAL)
    idx_x0x1 = _encode_multiindex(np.array([1, 1, 0, 0, 0, 0], dtype=np.int64), deg_p + deg_q, _ENCODE_DICT_GLOBAL)
    idx_x1sq = _encode_multiindex(np.array([0, 2, 0, 0, 0, 0], dtype=np.int64), deg_p + deg_q, _ENCODE_DICT_GLOBAL)
    
    assert abs(result[idx_x0sq] - 10.0) < 1e-10
    assert abs(result[idx_x0x1] - 27.0) < 1e-10  # 12 + 15
    assert abs(result[idx_x1sq] - 18.0) < 1e-10

def test_poly_mul_commutativity():
    """Test that polynomial multiplication is commutative: p * q = q * p"""
    deg_p = 2
    deg_q = 1
    
    # Create two polynomials with random coefficients
    p = _make_poly(deg_p, PSI)
    q = _make_poly(deg_q, PSI)
    
    # Set random coefficients
    for i in range(p.shape[0]):
        p[i] = i * 1.5
    
    for i in range(q.shape[0]):
        q[i] = i * 0.75
    
    # Compute p * q and q * p
    result1 = _poly_mul(p, deg_p, q, deg_q, PSI, CLMO, _ENCODE_DICT_GLOBAL) # Pass _ENCODE_DICT_GLOBAL
    result2 = _poly_mul(q, deg_q, p, deg_p, PSI, CLMO, _ENCODE_DICT_GLOBAL) # Pass _ENCODE_DICT_GLOBAL
    
    # They should be equal
    np.testing.assert_array_almost_equal(result1, result2)

def test_poly_mul_zero():
    """Test multiplication with zero polynomials"""
    deg_p = 2
    deg_q = 2
    
    # Create polynomials
    p = _make_poly(deg_p, PSI)
    zero_poly = _make_poly(deg_q, PSI)  # All zeros
    
    # Set some values in p
    for i in range(p.shape[0]):
        p[i] = i * 2.0
    
    # Multiply by zero
    result = _poly_mul(p, deg_p, zero_poly, deg_q, PSI, CLMO, _ENCODE_DICT_GLOBAL) # Pass _ENCODE_DICT_GLOBAL
    
    # Result should be all zeros
    assert np.all(result == 0.0)
    
    # Also test zero times anything is zero
    result = _poly_mul(zero_poly, deg_q, p, deg_p, PSI, CLMO, _ENCODE_DICT_GLOBAL) # Pass _ENCODE_DICT_GLOBAL
    assert np.all(result == 0.0)

def test_poly_mul_identity():
    """Test multiplication by the constant polynomial 1.0"""
    deg_p = 2
    
    # Constant polynomial is degree 0
    deg_const = 0
    
    # Create polynomials
    p = _make_poly(deg_p, PSI)
    const_one = _make_poly(deg_const, PSI)
    
    # Set const_one to be the constant polynomial 1.0
    const_one[0] = 1.0
    
    # Set some values in p
    for i in range(p.shape[0]):
        p[i] = i * 1.5
    
    # p * 1 should equal p
    result = _poly_mul(p, deg_p, const_one, deg_const, PSI, CLMO, _ENCODE_DICT_GLOBAL) # Pass _ENCODE_DICT_GLOBAL
    
    # The result should have the same degree as p
    assert result.shape[0] == PSI[N_VARS, deg_p]
    
    # The coefficients should match p
    np.testing.assert_array_almost_equal(result, p)
    
    # 1 * p should also equal p
    result = _poly_mul(const_one, deg_const, p, deg_p, PSI, CLMO, _ENCODE_DICT_GLOBAL) # Pass _ENCODE_DICT_GLOBAL
    np.testing.assert_array_almost_equal(result, p)

def test_poly_mul_complex():
    """Test multiplication of complex polynomials"""
    deg_p = 1
    deg_q = 1
    
    # Create complex polynomials
    p = _make_poly(deg_p, PSI)
    q = _make_poly(deg_q, PSI)
    
    # Set values in p and q
    # p = (1+i)*x0 + (2+2i)*x1
    idx_x0 = _encode_multiindex(np.array([1, 0, 0, 0, 0, 0], dtype=np.int64), deg_p, _ENCODE_DICT_GLOBAL) # Use _ENCODE_DICT_GLOBAL
    idx_x1 = _encode_multiindex(np.array([0, 1, 0, 0, 0, 0], dtype=np.int64), deg_p, _ENCODE_DICT_GLOBAL) # Use _ENCODE_DICT_GLOBAL
    
    p[idx_x0] = complex(1.0, 1.0)
    p[idx_x1] = complex(2.0, 2.0)
    
    # q = (3-i)*x0 + (1-i)*x1
    # idx_x0q and idx_x1q will be the same as idx_x0 and idx_x1 as deg_q = deg_p = 1
    q[idx_x0] = complex(3.0, -1.0) # Use idx_x0
    q[idx_x1] = complex(1.0, -1.0) # Use idx_x1
    
    # Multiply
    result = _poly_mul(p, deg_p, q, deg_q, PSI, CLMO, _ENCODE_DICT_GLOBAL) # Pass _ENCODE_DICT_GLOBAL
    
    # Expected result:
    # (1+i)*(3-i)*x0^2 = (3-i+3i-i^2)*x0^2 = (3-i+3i+1)*x0^2 = (4+2i)*x0^2
    # (1+i)*(1-i)*x0*x1 = (1-i+i-i^2)*x0*x1 = (1-i+i+1)*x0*x1 = 2*x0*x1
    # (2+2i)*(3-i)*x0*x1 = (6-2i+6i-2i^2)*x0*x1 = (6-2i+6i+2)*x0*x1 = (8+4i)*x0*x1
    # (2+2i)*(1-i)*x1^2 = (2-2i+2i-2i^2)*x1^2 = (2-2i+2i+2)*x1^2 = 4*x1^2
    
    # So combined terms:
    # (4+2i)*x0^2 + (10+4i)*x0*x1 + (4+0i)*x1^2
    
    idx_x0sq = _encode_multiindex(np.array([2, 0, 0, 0, 0, 0], dtype=np.int64), deg_p + deg_q, _ENCODE_DICT_GLOBAL) # Use _ENCODE_DICT_GLOBAL
    idx_x0x1 = _encode_multiindex(np.array([1, 1, 0, 0, 0, 0], dtype=np.int64), deg_p + deg_q, _ENCODE_DICT_GLOBAL) # Use _ENCODE_DICT_GLOBAL
    idx_x1sq = _encode_multiindex(np.array([0, 2, 0, 0, 0, 0], dtype=np.int64), deg_p + deg_q, _ENCODE_DICT_GLOBAL) # Use _ENCODE_DICT_GLOBAL
    
    # Check results
    assert abs(result[idx_x0sq].real - 4.0) < 1e-10
    assert abs(result[idx_x0sq].imag - 2.0) < 1e-10
    
    assert abs(result[idx_x0x1].real - 10.0) < 1e-10
    assert abs(result[idx_x0x1].imag - 4.0) < 1e-10
    
    assert abs(result[idx_x1sq].real - 4.0) < 1e-10
    assert abs(result[idx_x1sq].imag) < 1e-10

def test_differentiate_monomial():
    """Test differentiation of a monomial"""
    # Test differentiation of a simple monomial: 3*x0^2
    degree = 2
    p = _make_poly(degree, PSI)
    
    # Set p = 3*x0^2
    idx = _encode_multiindex(np.array([2, 0, 0, 0, 0, 0], dtype=np.int64), degree, _ENCODE_DICT_GLOBAL) # Use _ENCODE_DICT_GLOBAL
    p[idx] = 3.0
    
    # Differentiate with respect to x0
    dp = _poly_diff(p, 0, degree, PSI, CLMO, _ENCODE_DICT_GLOBAL) # Pass _ENCODE_DICT_GLOBAL
    
    # Result should be 6*x0 (d/dx0(3*x0^2) = 6*x0)
    idx_x0 = _encode_multiindex(np.array([1, 0, 0, 0, 0, 0], dtype=np.int64), degree-1, _ENCODE_DICT_GLOBAL) # Use _ENCODE_DICT_GLOBAL
    assert dp[idx_x0] == 6.0
    
    # All other terms should be zero
    dp[idx_x0] = 0.0
    assert np.all(dp == 0.0)
    
    # Differentiate with respect to a variable that doesn't appear in the term
    dp2 = _poly_diff(p, 1, degree, PSI, CLMO, _ENCODE_DICT_GLOBAL) # Pass _ENCODE_DICT_GLOBAL
    
    # Result should be all zeros
    assert np.all(dp2 == 0.0)

def test_differentiate_mixed_terms():
    """Test differentiation of a polynomial with multiple terms"""
    # Create p = 2*x0^2 + 3*x0*x1 + 4*x1^2
    degree = 2
    p = _make_poly(degree, PSI)
    
    idx_x0sq = _encode_multiindex(np.array([2, 0, 0, 0, 0, 0], dtype=np.int64), degree, _ENCODE_DICT_GLOBAL) # Use _ENCODE_DICT_GLOBAL
    idx_x0x1 = _encode_multiindex(np.array([1, 1, 0, 0, 0, 0], dtype=np.int64), degree, _ENCODE_DICT_GLOBAL) # Use _ENCODE_DICT_GLOBAL
    idx_x1sq = _encode_multiindex(np.array([0, 2, 0, 0, 0, 0], dtype=np.int64), degree, _ENCODE_DICT_GLOBAL) # Use _ENCODE_DICT_GLOBAL
    
    p[idx_x0sq] = 2.0
    p[idx_x0x1] = 3.0
    p[idx_x1sq] = 4.0
    
    # Differentiate with respect to x0
    dp = _poly_diff(p, 0, degree, PSI, CLMO, _ENCODE_DICT_GLOBAL) # Pass _ENCODE_DICT_GLOBAL
    
    # d/dx0(2*x0^2 + 3*x0*x1 + 4*x1^2) = 4*x0 + 3*x1
    idx_x0 = _encode_multiindex(np.array([1, 0, 0, 0, 0, 0], dtype=np.int64), degree-1, _ENCODE_DICT_GLOBAL) # Use _ENCODE_DICT_GLOBAL
    idx_x1 = _encode_multiindex(np.array([0, 1, 0, 0, 0, 0], dtype=np.int64), degree-1, _ENCODE_DICT_GLOBAL) # Use _ENCODE_DICT_GLOBAL
    
    assert dp[idx_x0] == 4.0
    assert dp[idx_x1] == 3.0
    
    # All other terms should be zero
    dp[idx_x0] = 0.0
    dp[idx_x1] = 0.0
    assert np.all(dp == 0.0)
    
    # Differentiate with respect to x1
    dp2 = _poly_diff(p, 1, degree, PSI, CLMO, _ENCODE_DICT_GLOBAL) # Pass _ENCODE_DICT_GLOBAL
    
    # d/dx1(2*x0^2 + 3*x0*x1 + 4*x1^2) = 3*x0 + 8*x1
    idx_x0 = _encode_multiindex(np.array([1, 0, 0, 0, 0, 0], dtype=np.int64), degree-1, _ENCODE_DICT_GLOBAL) # Use _ENCODE_DICT_GLOBAL
    idx_x1 = _encode_multiindex(np.array([0, 1, 0, 0, 0, 0], dtype=np.int64), degree-1, _ENCODE_DICT_GLOBAL) # Use _ENCODE_DICT_GLOBAL
    
    assert dp2[idx_x0] == 3.0
    assert dp2[idx_x1] == 8.0
    
    # All other terms should be zero
    dp2[idx_x0] = 0.0
    dp2[idx_x1] = 0.0
    assert np.all(dp2 == 0.0)

def test_differentiate_higher_degree():
    """Test differentiation of higher-degree polynomials"""
    # Create p = x0^3 + x0^2*x1 + x0*x1^2 + x1^3
    degree = 3
    p = _make_poly(degree, PSI)
    
    idx_x0_3 = _encode_multiindex(np.array([3, 0, 0, 0, 0, 0], dtype=np.int64), degree, _ENCODE_DICT_GLOBAL) # Use _ENCODE_DICT_GLOBAL
    idx_x0_2_x1 = _encode_multiindex(np.array([2, 1, 0, 0, 0, 0], dtype=np.int64), degree, _ENCODE_DICT_GLOBAL) # Use _ENCODE_DICT_GLOBAL
    idx_x0_x1_2 = _encode_multiindex(np.array([1, 2, 0, 0, 0, 0], dtype=np.int64), degree, _ENCODE_DICT_GLOBAL) # Use _ENCODE_DICT_GLOBAL
    idx_x1_3 = _encode_multiindex(np.array([0, 3, 0, 0, 0, 0], dtype=np.int64), degree, _ENCODE_DICT_GLOBAL) # Use _ENCODE_DICT_GLOBAL
    
    p[idx_x0_3] = 1.0
    p[idx_x0_2_x1] = 1.0
    p[idx_x0_x1_2] = 1.0
    p[idx_x1_3] = 1.0
    
    # Differentiate with respect to x0
    dp = _poly_diff(p, 0, degree, PSI, CLMO, _ENCODE_DICT_GLOBAL) # Pass _ENCODE_DICT_GLOBAL
    
    # d/dx0(x0^3 + x0^2*x1 + x0*x1^2 + x1^3) = 3*x0^2 + 2*x0*x1 + x1^2
    idx_x0sq = _encode_multiindex(np.array([2, 0, 0, 0, 0, 0], dtype=np.int64), degree-1, _ENCODE_DICT_GLOBAL) # Use _ENCODE_DICT_GLOBAL
    idx_x0x1 = _encode_multiindex(np.array([1, 1, 0, 0, 0, 0], dtype=np.int64), degree-1, _ENCODE_DICT_GLOBAL) # Use _ENCODE_DICT_GLOBAL
    idx_x1sq = _encode_multiindex(np.array([0, 2, 0, 0, 0, 0], dtype=np.int64), degree-1, _ENCODE_DICT_GLOBAL) # Use _ENCODE_DICT_GLOBAL
    
    assert dp[idx_x0sq] == 3.0
    assert dp[idx_x0x1] == 2.0
    assert dp[idx_x1sq] == 1.0
    
    # Check all other terms are zero
    dp[idx_x0sq] = 0.0
    dp[idx_x0x1] = 0.0
    dp[idx_x1sq] = 0.0
    assert np.all(dp == 0.0)

def test_differentiate_zero_polynomial():
    """Test differentiation of zero polynomial"""
    degree = 2
    p = _make_poly(degree, PSI)  # All zeros by default
    
    # Differentiate zero polynomial
    dp = _poly_diff(p, 0, degree, PSI, CLMO, _ENCODE_DICT_GLOBAL) # Pass _ENCODE_DICT_GLOBAL
    
    # Result should be a zero polynomial of degree-1
    assert np.all(dp == 0.0)

def test_differentiate_constant():
    """Test differentiation of a constant term (should be zero)"""
    # Note: In this implementation, a degree-0 polynomial is just a constant
    degree = 0
    p = _make_poly(degree, PSI)
    p[0] = 5.0  # Constant polynomial with value 5
    
    # For a Numba JIT function, trying to _poly_diff a constant will not raise an exception
    # but should return an appropriate result (likely an empty array or zeros)
    for var in range(N_VARS):
        dp = _poly_diff(p, var, degree, PSI, CLMO, _ENCODE_DICT_GLOBAL) # Pass _ENCODE_DICT_GLOBAL
        
        # The size should be PSI[N_VARS, -1] which should be 0 or a very small array
        # Just verify that the result doesn't contain any non-zero values
        for i in range(dp.shape[0]):
            assert dp[i] == 0.0

def test_differentiate_multi_variable():
    """Test differentiation with respect to different variables"""
    # Create p = x0*x1*x2
    degree = 3
    p = _make_poly(degree, PSI)
    
    idx = _encode_multiindex(np.array([1, 1, 1, 0, 0, 0], dtype=np.int64), degree, _ENCODE_DICT_GLOBAL) # Use _ENCODE_DICT_GLOBAL
    p[idx] = 2.0
    
    # Differentiate with respect to x0
    dp0 = _poly_diff(p, 0, degree, PSI, CLMO, _ENCODE_DICT_GLOBAL) # Pass _ENCODE_DICT_GLOBAL
    
    # d/dx0(2*x0*x1*x2) = 2*x1*x2
    idx_x1x2 = _encode_multiindex(np.array([0, 1, 1, 0, 0, 0], dtype=np.int64), degree-1, _ENCODE_DICT_GLOBAL) # Use _ENCODE_DICT_GLOBAL
    assert dp0[idx_x1x2] == 2.0
    
    # Differentiate with respect to x1
    dp1 = _poly_diff(p, 1, degree, PSI, CLMO, _ENCODE_DICT_GLOBAL) # Pass _ENCODE_DICT_GLOBAL
    
    # d/dx1(2*x0*x1*x2) = 2*x0*x2
    idx_x0x2 = _encode_multiindex(np.array([1, 0, 1, 0, 0, 0], dtype=np.int64), degree-1, _ENCODE_DICT_GLOBAL) # Use _ENCODE_DICT_GLOBAL
    assert dp1[idx_x0x2] == 2.0
    
    # Differentiate with respect to x2
    dp2 = _poly_diff(p, 2, degree, PSI, CLMO, _ENCODE_DICT_GLOBAL) # Pass _ENCODE_DICT_GLOBAL
    
    # d/dx2(2*x0*x1*x2) = 2*x0*x1
    idx_x0x1 = _encode_multiindex(np.array([1, 1, 0, 0, 0, 0], dtype=np.int64), degree-1, _ENCODE_DICT_GLOBAL) # Use _ENCODE_DICT_GLOBAL
    assert dp2[idx_x0x1] == 2.0
    
    # Differentiate with respect to x3 (not in the polynomial)
    dp3 = _poly_diff(p, 3, degree, PSI, CLMO, _ENCODE_DICT_GLOBAL) # Pass _ENCODE_DICT_GLOBAL
    
    # d/dx3(2*x0*x1*x2) = 0
    assert np.all(dp3 == 0.0)

def test_differentiate_high_exponent():
    """Test differentiation of terms with high exponents"""
    # Create p = x0^4
    degree = 4
    p = _make_poly(degree, PSI)
    
    idx = _encode_multiindex(np.array([4, 0, 0, 0, 0, 0], dtype=np.int64), degree, _ENCODE_DICT_GLOBAL) # Use _ENCODE_DICT_GLOBAL
    p[idx] = 1.0
    
    # Differentiate with respect to x0
    dp = _poly_diff(p, 0, degree, PSI, CLMO, _ENCODE_DICT_GLOBAL) # Pass _ENCODE_DICT_GLOBAL
    
    # d/dx0(x0^4) = 4*x0^3
    idx_x0_3 = _encode_multiindex(np.array([3, 0, 0, 0, 0, 0], dtype=np.int64), degree-1, _ENCODE_DICT_GLOBAL) # Use _ENCODE_DICT_GLOBAL
    assert dp[idx_x0_3] == 4.0
    
    # Differentiate again
    dp2 = _poly_diff(dp, 0, degree-1, PSI, CLMO, _ENCODE_DICT_GLOBAL) # Pass _ENCODE_DICT_GLOBAL
    
    # d/dx0(4*x0^3) = 12*x0^2
    idx_x0sq = _encode_multiindex(np.array([2, 0, 0, 0, 0, 0], dtype=np.int64), degree-2, _ENCODE_DICT_GLOBAL) # Use _ENCODE_DICT_GLOBAL
    assert dp2[idx_x0sq] == 12.0
    
    # Differentiate a third time
    dp3 = _poly_diff(dp2, 0, degree-2, PSI, CLMO, _ENCODE_DICT_GLOBAL) # Pass _ENCODE_DICT_GLOBAL
    
    # d/dx0(12*x0^2) = 24*x0
    idx_x0 = _encode_multiindex(np.array([1, 0, 0, 0, 0, 0], dtype=np.int64), degree-3, _ENCODE_DICT_GLOBAL) # Use _ENCODE_DICT_GLOBAL
    assert dp3[idx_x0] == 24.0
    
    # Differentiate a fourth time
    dp4 = _poly_diff(dp3, 0, degree-3, PSI, CLMO, _ENCODE_DICT_GLOBAL) # Pass _ENCODE_DICT_GLOBAL
    
    # d/dx0(24*x0) = 24
    idx_const = _encode_multiindex(np.array([0, 0, 0, 0, 0, 0], dtype=np.int64), degree-4, _ENCODE_DICT_GLOBAL) # Use _ENCODE_DICT_GLOBAL
    assert dp4[idx_const] == 24.0

def test_differentiate_complex():
    """Test differentiation of complex polynomials"""
    # Create a complex polynomial
    degree = 2
    p = _make_poly(degree, PSI)
    
    # Set p = (1+2i)*x0^2 + (3+4i)*x0*x1 + (5+6i)*x1^2
    idx_x0sq = _encode_multiindex(np.array([2, 0, 0, 0, 0, 0], dtype=np.int64), degree, _ENCODE_DICT_GLOBAL) # Use _ENCODE_DICT_GLOBAL
    idx_x0x1 = _encode_multiindex(np.array([1, 1, 0, 0, 0, 0], dtype=np.int64), degree, _ENCODE_DICT_GLOBAL) # Use _ENCODE_DICT_GLOBAL
    idx_x1sq = _encode_multiindex(np.array([0, 2, 0, 0, 0, 0], dtype=np.int64), degree, _ENCODE_DICT_GLOBAL) # Use _ENCODE_DICT_GLOBAL
    
    p[idx_x0sq] = complex(1.0, 2.0)
    p[idx_x0x1] = complex(3.0, 4.0)
    p[idx_x1sq] = complex(5.0, 6.0)
    
    # Differentiate with respect to x0
    dp = _poly_diff(p, 0, degree, PSI, CLMO, _ENCODE_DICT_GLOBAL)
    
    # d/dx0((1+2i)*x0^2 + (3+4i)*x0*x1 + (5+6i)*x1^2) = (2+4i)*x0 + (3+4i)*x1
    idx_x0 = _encode_multiindex(np.array([1, 0, 0, 0, 0, 0], dtype=np.int64), degree-1, _ENCODE_DICT_GLOBAL) # Use _ENCODE_DICT_GLOBAL
    idx_x1 = _encode_multiindex(np.array([0, 1, 0, 0, 0, 0], dtype=np.int64), degree-1, _ENCODE_DICT_GLOBAL) # Use _ENCODE_DICT_GLOBAL
    
    assert dp[idx_x0].real == 2.0
    assert dp[idx_x0].imag == 4.0
    assert dp[idx_x1].real == 3.0
    assert dp[idx_x1].imag == 4.0
    
    # Differentiate with respect to x1
    dp2 = _poly_diff(p, 1, degree, PSI, CLMO, _ENCODE_DICT_GLOBAL)
    
    # d/dx1((1+2i)*x0^2 + (3+4i)*x0*x1 + (5+6i)*x1^2) = (3+4i)*x0 + (10+12i)*x1
    idx_x0 = _encode_multiindex(np.array([1, 0, 0, 0, 0, 0], dtype=np.int64), degree-1, _ENCODE_DICT_GLOBAL) # Use _ENCODE_DICT_GLOBAL
    idx_x1 = _encode_multiindex(np.array([0, 1, 0, 0, 0, 0], dtype=np.int64), degree-1, _ENCODE_DICT_GLOBAL) # Use _ENCODE_DICT_GLOBAL
    
    assert dp2[idx_x0].real == 3.0
    assert dp2[idx_x0].imag == 4.0
    assert dp2[idx_x1].real == 10.0
    assert dp2[idx_x1].imag == 12.0

def test_poisson_antisymmetry():
    """Test antisymmetry property of _poly_poisson bracket: {P, Q} = -{Q, P}"""
    # Create two test polynomials
    deg_p = 2
    deg_q = 2
    
    p = _make_poly(deg_p, PSI)
    q = _make_poly(deg_q, PSI)
    
    # Set values in p and q
    # p = x0^2 + x0*x1
    idx_x0sq = _encode_multiindex(np.array([2, 0, 0, 0, 0, 0], dtype=np.int64), deg_p, _ENCODE_DICT_GLOBAL) # Use _ENCODE_DICT_GLOBAL
    idx_x0x1 = _encode_multiindex(np.array([1, 1, 0, 0, 0, 0], dtype=np.int64), deg_p, _ENCODE_DICT_GLOBAL) # Use _ENCODE_DICT_GLOBAL
    
    p[idx_x0sq] = 1.0
    p[idx_x0x1] = 1.0
    
    # q = p0^2 + p0*p1
    idx_p0sq = _encode_multiindex(np.array([0, 0, 0, 2, 0, 0], dtype=np.int64), deg_q, _ENCODE_DICT_GLOBAL) # Use _ENCODE_DICT_GLOBAL
    idx_p0p1 = _encode_multiindex(np.array([0, 0, 0, 1, 1, 0], dtype=np.int64), deg_q, _ENCODE_DICT_GLOBAL) # Use _ENCODE_DICT_GLOBAL
    
    q[idx_p0sq] = 1.0
    q[idx_p0p1] = 1.0
    
    # Compute {P, Q} and {Q, P}
    pq = _poly_poisson(p, deg_p, q, deg_q, PSI, CLMO, _ENCODE_DICT_GLOBAL)
    qp = _poly_poisson(q, deg_q, p, deg_p, PSI, CLMO, _ENCODE_DICT_GLOBAL)
    
    # Compute -{Q, P} (scale qp by -1)
    neg_qp = np.zeros_like(qp)
    _poly_scale(qp, -1.0, neg_qp)
    
    # {P, Q} should equal -{Q, P}
    np.testing.assert_array_almost_equal(pq, neg_qp)

def test_poisson_first_argument_linearity():
    """Test linearity in the first argument: {aP+bQ, R} = a{P,R} + b{Q,R}"""
    # Create three test polynomials
    deg_p = 2
    deg_q = 2
    deg_r = 2
    
    p = _make_poly(deg_p, PSI)
    q = _make_poly(deg_q, PSI)
    r = _make_poly(deg_r, PSI)
    
    # Set values
    # p = x0^2
    idx_x0sq = _encode_multiindex(np.array([2, 0, 0, 0, 0, 0], dtype=np.int64), deg_p, _ENCODE_DICT_GLOBAL)
    p[idx_x0sq] = 1.0
    
    # q = x1^2
    idx_x1sq = _encode_multiindex(np.array([0, 2, 0, 0, 0, 0], dtype=np.int64), deg_q, _ENCODE_DICT_GLOBAL)
    q[idx_x1sq] = 1.0
    
    # r = p0^2
    idx_p0sq = _encode_multiindex(np.array([0, 0, 0, 2, 0, 0], dtype=np.int64), deg_r, _ENCODE_DICT_GLOBAL)
    r[idx_p0sq] = 1.0
    
    # Define coefficients a and b
    a = 2.0
    b = 3.0
    
    # Compute a*p + b*q
    ap = _make_poly(deg_p, PSI)
    bq = _make_poly(deg_q, PSI)
    apbq = _make_poly(deg_p, PSI)  # assuming deg_p == deg_q
    
    _poly_scale(p, a, ap)
    _poly_scale(q, b, bq)
    _poly_add(ap, bq, apbq)
    
    # Compute {aP+bQ, R}
    apbq_r = _poly_poisson(apbq, deg_p, r, deg_r, PSI, CLMO, _ENCODE_DICT_GLOBAL)
    
    # Compute a{P,R} + b{Q,R}
    pr = _poly_poisson(p, deg_p, r, deg_r, PSI, CLMO, _ENCODE_DICT_GLOBAL)
    qr = _poly_poisson(q, deg_q, r, deg_r, PSI, CLMO, _ENCODE_DICT_GLOBAL)
    
    apr = _make_poly(deg_p + deg_r - 2, PSI)
    bqr = _make_poly(deg_q + deg_r - 2, PSI)
    result = _make_poly(deg_p + deg_r - 2, PSI)
    
    _poly_scale(pr, a, apr)
    _poly_scale(qr, b, bqr)
    _poly_add(apr, bqr, result)
    
    # {aP+bQ, R} should equal a{P,R} + b{Q,R}
    np.testing.assert_array_almost_equal(apbq_r, result)

def test_poisson_second_argument_linearity():
    """Test linearity in the second argument: {P, aQ+bR} = a{P,Q} + b{P,R}"""
    # Create three test polynomials
    deg_p = 2
    deg_q = 2
    deg_r = 2
    
    p = _make_poly(deg_p, PSI)
    q = _make_poly(deg_q, PSI)
    r = _make_poly(deg_r, PSI)
    
    # Set values
    # p = x0^2
    idx_x0sq = _encode_multiindex(np.array([2, 0, 0, 0, 0, 0], dtype=np.int64), deg_p, _ENCODE_DICT_GLOBAL)
    p[idx_x0sq] = 1.0
    
    # q = x1^2
    idx_x1sq = _encode_multiindex(np.array([0, 2, 0, 0, 0, 0], dtype=np.int64), deg_q, _ENCODE_DICT_GLOBAL)
    q[idx_x1sq] = 1.0
    
    # r = p0^2
    idx_p0sq = _encode_multiindex(np.array([0, 0, 0, 2, 0, 0], dtype=np.int64), deg_r, _ENCODE_DICT_GLOBAL)
    r[idx_p0sq] = 1.0
    
    # Define coefficients a and b
    a = 2.0
    b = 3.0
    
    # Compute a*q + b*r
    aq = _make_poly(deg_q, PSI)
    br = _make_poly(deg_r, PSI)
    aqbr = _make_poly(deg_q, PSI)  # assuming deg_q == deg_r
    
    _poly_scale(q, a, aq)
    _poly_scale(r, b, br)
    _poly_add(aq, br, aqbr)
    
    # Compute {P, aQ+bR}
    p_aqbr = _poly_poisson(p, deg_p, aqbr, deg_q, PSI, CLMO, _ENCODE_DICT_GLOBAL)
    
    # Compute a{P,Q} + b{P,R}
    pq = _poly_poisson(p, deg_p, q, deg_q, PSI, CLMO, _ENCODE_DICT_GLOBAL)
    pr = _poly_poisson(p, deg_p, r, deg_r, PSI, CLMO, _ENCODE_DICT_GLOBAL)
    
    apq = _make_poly(deg_p + deg_q - 2, PSI)
    bpr = _make_poly(deg_p + deg_r - 2, PSI)
    result = _make_poly(deg_p + deg_q - 2, PSI)
    
    _poly_scale(pq, a, apq)
    _poly_scale(pr, b, bpr)
    _poly_add(apq, bpr, result)
    
    # {P, aQ+bR} should equal a{P,Q} + b{P,R}
    np.testing.assert_array_almost_equal(p_aqbr, result)

def test_poisson_jacobi_identity():
    """Test Jacobi identity: {P, {Q, R}} + {Q, {R, P}} + {R, {P, Q}} = 0"""
    # Create three test polynomials of low degree to keep test manageable
    deg_p = 2
    deg_q = 2
    deg_r = 2
    
    p = _make_poly(deg_p, PSI)
    q = _make_poly(deg_q, PSI)
    r = _make_poly(deg_r, PSI)
    
    # Set values - simple polynomials to test the identity
    # p = x0^2
    idx_x0sq = _encode_multiindex(np.array([2, 0, 0, 0, 0, 0], dtype=np.int64), deg_p, _ENCODE_DICT_GLOBAL)
    p[idx_x0sq] = 1.0
    
    # q = p0^2
    idx_p0sq = _encode_multiindex(np.array([0, 0, 0, 2, 0, 0], dtype=np.int64), deg_q, _ENCODE_DICT_GLOBAL)
    q[idx_p0sq] = 1.0
    
    # r = x0*p0
    idx_x0p0 = _encode_multiindex(np.array([1, 0, 0, 1, 0, 0], dtype=np.int64), deg_r, _ENCODE_DICT_GLOBAL)
    r[idx_x0p0] = 1.0
    
    # Compute {Q, R}
    qr_deg = deg_q + deg_r - 2
    qr = _poly_poisson(q, deg_q, r, deg_r, PSI, CLMO, _ENCODE_DICT_GLOBAL)
    
    # Compute {P, {Q, R}}
    p_qr_deg = deg_p + qr_deg - 2
    p_qr = _poly_poisson(p, deg_p, qr, qr_deg, PSI, CLMO, _ENCODE_DICT_GLOBAL)
    
    # Compute {R, P}
    rp_deg = deg_r + deg_p - 2
    rp = _poly_poisson(r, deg_r, p, deg_p, PSI, CLMO, _ENCODE_DICT_GLOBAL)
    
    # Compute {Q, {R, P}}
    q_rp_deg = deg_q + rp_deg - 2
    q_rp = _poly_poisson(q, deg_q, rp, rp_deg, PSI, CLMO, _ENCODE_DICT_GLOBAL)
    
    # Compute {P, Q}
    pq_deg = deg_p + deg_q - 2
    pq = _poly_poisson(p, deg_p, q, deg_q, PSI, CLMO, _ENCODE_DICT_GLOBAL)
    
    # Compute {R, {P, Q}}
    r_pq_deg = deg_r + pq_deg - 2
    r_pq = _poly_poisson(r, deg_r, pq, pq_deg, PSI, CLMO, _ENCODE_DICT_GLOBAL)
    
    # Compute the sum: {P, {Q, R}} + {Q, {R, P}} + {R, {P, Q}}
    sum_deg = max(p_qr_deg, q_rp_deg, r_pq_deg)
    sum_result = _make_poly(sum_deg, PSI)
    
    temp = _make_poly(sum_deg, PSI)
    _poly_add(p_qr, q_rp, temp)
    _poly_add(temp, r_pq, sum_result)
    
    # The sum should be zero
    for i in range(sum_result.shape[0]):
        assert abs(sum_result[i]) < 1e-10

def test_poisson_hamiltonian():
    """Test _poly_poisson bracket properties with the Hamiltonian H = 0.5*p0^2 + q0^2"""
    # Create Hamiltonian H = 0.5*p0^2 + q0^2
    deg_h = 2
    h = _make_poly(deg_h, PSI)
    
    # 0.5*p0^2 term
    idx_p0sq = _encode_multiindex(np.array([0, 0, 0, 2, 0, 0], dtype=np.int64), deg_h, _ENCODE_DICT_GLOBAL)
    h[idx_p0sq] = 0.5
    
    # q0^2 term (x0^2)
    idx_x0sq = _encode_multiindex(np.array([2, 0, 0, 0, 0, 0], dtype=np.int64), deg_h, _ENCODE_DICT_GLOBAL)
    h[idx_x0sq] = 1.0
    
    # Test 1: {H, H} = 0
    hh = _poly_poisson(h, deg_h, h, deg_h, PSI, CLMO, _ENCODE_DICT_GLOBAL)
    
    # {H, H} should be zero
    for i in range(hh.shape[0]):
        assert abs(hh[i]) < 1e-10
    
    # Test 2: {H, q0} = -p0
    # Create q0 polynomial (x0)
    deg_q0 = 1
    q0 = _make_poly(deg_q0, PSI)
    idx_x0 = _encode_multiindex(np.array([1, 0, 0, 0, 0, 0], dtype=np.int64), deg_q0, _ENCODE_DICT_GLOBAL)
    q0[idx_x0] = 1.0
    
    # Compute {H, q0}
    h_q0 = _poly_poisson(h, deg_h, q0, deg_q0, PSI, CLMO, _ENCODE_DICT_GLOBAL)
    
    # Expected result: -p0
    idx_p0 = _encode_multiindex(np.array([0, 0, 0, 1, 0, 0], dtype=np.int64), 1, _ENCODE_DICT_GLOBAL)
    expected_h_q0 = _make_poly(1, PSI)
    expected_h_q0[idx_p0] = -1.0
    
    # {H, q0} should equal -p0
    np.testing.assert_array_almost_equal(h_q0, expected_h_q0)
    
    # Test 3: {H, p0} = 2*q0 (for this implementation)
    # Create p0 polynomial
    deg_p0 = 1
    p0 = _make_poly(deg_p0, PSI)
    idx_p0 = _encode_multiindex(np.array([0, 0, 0, 1, 0, 0], dtype=np.int64), deg_p0, _ENCODE_DICT_GLOBAL)
    p0[idx_p0] = 1.0
    
    # Compute {H, p0}
    h_p0 = _poly_poisson(h, deg_h, p0, deg_p0, PSI, CLMO, _ENCODE_DICT_GLOBAL)
    
    # Expected result: 2*q0 (2*x0)
    idx_x0 = _encode_multiindex(np.array([1, 0, 0, 0, 0, 0], dtype=np.int64), 1, _ENCODE_DICT_GLOBAL)
    expected_h_p0 = _make_poly(1, PSI)
    expected_h_p0[idx_x0] = 2.0
    
    # {H, p0} should equal 2*q0
    np.testing.assert_array_almost_equal(h_p0, expected_h_p0)

def test_poisson_complex():
    """Test _poly_poisson bracket with complex polynomials"""
    # Create complex polynomials
    deg_p = 2
    deg_q = 2
    
    p = _make_poly(deg_p, PSI)
    q = _make_poly(deg_q, PSI)
    
    # Set values in p and q
    # p = (1+i)*x0^2 + (2+2i)*p0^2
    idx_x0sq = _encode_multiindex(np.array([2, 0, 0, 0, 0, 0], dtype=np.int64), deg_p, _ENCODE_DICT_GLOBAL)
    idx_p0sq = _encode_multiindex(np.array([0, 0, 0, 2, 0, 0], dtype=np.int64), deg_p, _ENCODE_DICT_GLOBAL)
    
    p[idx_x0sq] = complex(1.0, 1.0)
    p[idx_p0sq] = complex(2.0, 2.0)
    
    # q = (3-i)*x0*p0
    idx_x0p0 = _encode_multiindex(np.array([1, 0, 0, 1, 0, 0], dtype=np.int64), deg_q, _ENCODE_DICT_GLOBAL)
    q[idx_x0p0] = complex(3.0, -1.0)
    
    # Compute {P, Q} and {Q, P}
    pq = _poly_poisson(p, deg_p, q, deg_q, PSI, CLMO, _ENCODE_DICT_GLOBAL)
    qp = _poly_poisson(q, deg_q, p, deg_p, PSI, CLMO, _ENCODE_DICT_GLOBAL)
    
    # Test antisymmetry with complex polynomials
    neg_qp = _make_poly(deg_p + deg_q - 2, PSI)
    _poly_scale(qp, -1.0, neg_qp)
    
    # {P, Q} should equal -{Q, P}
    np.testing.assert_array_almost_equal(pq, neg_qp)
    
    # Verify a specific term in the result to check complex arithmetic correctness
    # We'd need to calculate the expected result manually for a specific term

def test_poisson_leibniz_rule():
    """Test Leibniz product rule: {P, Q*R} = {P, Q}*R + Q*{P, R}"""
    # Create test polynomials of low degree
    deg_p = 2
    deg_q = 1
    deg_r = 1
    
    p = _make_poly(deg_p, PSI)
    q = _make_poly(deg_q, PSI)
    r = _make_poly(deg_r, PSI)
    
    # Set values for simple test case
    # p = x0^2
    idx_x0sq = _encode_multiindex(np.array([2, 0, 0, 0, 0, 0], dtype=np.int64), deg_p, _ENCODE_DICT_GLOBAL)
    p[idx_x0sq] = 1.0
    
    # q = x1
    idx_x1 = _encode_multiindex(np.array([0, 1, 0, 0, 0, 0], dtype=np.int64), deg_q, _ENCODE_DICT_GLOBAL)
    q[idx_x1] = 1.0
    
    # r = p0
    idx_p0 = _encode_multiindex(np.array([0, 0, 0, 1, 0, 0], dtype=np.int64), deg_r, _ENCODE_DICT_GLOBAL)
    r[idx_p0] = 1.0
    
    # Compute Q*R
    qr = _poly_mul(q, deg_q, r, deg_r, PSI, CLMO, _ENCODE_DICT_GLOBAL)
    deg_qr = deg_q + deg_r
    
    # Compute {P, Q*R}
    p_qr = _poly_poisson(p, deg_p, qr, deg_qr, PSI, CLMO, _ENCODE_DICT_GLOBAL)
    
    # Compute {P, Q}
    pq = _poly_poisson(p, deg_p, q, deg_q, PSI, CLMO, _ENCODE_DICT_GLOBAL)
    
    # Compute {P, Q}*R
    pq_r = _poly_mul(pq, deg_p + deg_q - 2, r, deg_r, PSI, CLMO, _ENCODE_DICT_GLOBAL)
    
    # Compute {P, R}
    pr = _poly_poisson(p, deg_p, r, deg_r, PSI, CLMO, _ENCODE_DICT_GLOBAL)
    
    # Compute Q*{P, R}
    q_pr = _poly_mul(q, deg_q, pr, deg_p + deg_r - 2, PSI, CLMO, _ENCODE_DICT_GLOBAL)
    
    # Compute {P, Q}*R + Q*{P, R}
    result = _make_poly(deg_p + deg_qr - 2, PSI)
    _poly_add(pq_r, q_pr, result)
    
    # {P, Q*R} should equal {P, Q}*R + Q*{P, R}
    np.testing.assert_array_almost_equal(p_qr, result)

def test_poisson_constant():
    """Test _poly_poisson bracket with constant function: {1, F} = 0"""
    # Create constant polynomial 1
    deg_const = 0
    const_one = _make_poly(deg_const, PSI)
    const_one[0] = 1.0
    
    # Create test polynomial
    deg_f = 2
    f = _make_poly(deg_f, PSI)
    
    # Set f = x0^2 + p0^2
    idx_x0sq = _encode_multiindex(np.array([2, 0, 0, 0, 0, 0], dtype=np.int64), deg_f, _ENCODE_DICT_GLOBAL)
    idx_p0sq = _encode_multiindex(np.array([0, 0, 0, 2, 0, 0], dtype=np.int64), deg_f, _ENCODE_DICT_GLOBAL)
    
    f[idx_x0sq] = 1.0
    f[idx_p0sq] = 1.0
    
    # Compute {1, F}
    one_f = _poly_poisson(const_one, deg_const, f, deg_f, PSI, CLMO, _ENCODE_DICT_GLOBAL)
    
    # {1, F} should be zero
    for i in range(one_f.shape[0]):
        assert abs(one_f[i]) < 1e-10
    
    # Test the other way: {F, 1}
    f_one = _poly_poisson(f, deg_f, const_one, deg_const, PSI, CLMO, _ENCODE_DICT_GLOBAL)
    
    # {F, 1} should also be zero
    for i in range(f_one.shape[0]):
        assert abs(f_one[i]) < 1e-10

def test_poisson_canonical_relations():
    """Test canonical _poly_poisson bracket relations: {q_i,q_j}=0, {p_i,p_j}=0, {q_i,p_j}=delta _ij"""
    # Test for the first few position and momentum variables
    for i in range(3):  # Test x0, x1, x2
        for j in range(3):  # Test x0, x1, x2
            # Create position variables q_i and q_j
            deg_q = 1
            qi = _make_poly(deg_q, PSI)
            qj = _make_poly(deg_q, PSI)
            
            # Create arrays for exponents first, then encode them
            idx_qi_array = np.array([0, 0, 0, 0, 0, 0], dtype=np.int64)
            idx_qi_array[i] = 1  # Set exponent for variable i
            
            idx_qj_array = np.array([0, 0, 0, 0, 0, 0], dtype=np.int64)
            idx_qj_array[j] = 1  # Set exponent for variable j
            
            qi_idx = _encode_multiindex(idx_qi_array, deg_q, _ENCODE_DICT_GLOBAL)
            qj_idx = _encode_multiindex(idx_qj_array, deg_q, _ENCODE_DICT_GLOBAL)
            
            qi[qi_idx] = 1.0
            qj[qj_idx] = 1.0
            
            # Create momentum variables p_i and p_j
            pi = _make_poly(deg_q, PSI)
            pj = _make_poly(deg_q, PSI)
            
            # Create arrays for exponents first, then encode them
            idx_pi_array = np.array([0, 0, 0, 0, 0, 0], dtype=np.int64)
            idx_pi_array[i+3] = 1  # p0,p1,p2 are at indices 3,4,5
            
            idx_pj_array = np.array([0, 0, 0, 0, 0, 0], dtype=np.int64)
            idx_pj_array[j+3] = 1  # p0,p1,p2 are at indices 3,4,5
            
            pi_idx = _encode_multiindex(idx_pi_array, deg_q, _ENCODE_DICT_GLOBAL)
            pj_idx = _encode_multiindex(idx_pj_array, deg_q, _ENCODE_DICT_GLOBAL)
            
            pi[pi_idx] = 1.0
            pj[pj_idx] = 1.0
            
            # Test {q_i, q_j} = 0
            qi_qj = _poly_poisson(qi, deg_q, qj, deg_q, PSI, CLMO, _ENCODE_DICT_GLOBAL)
            
            # Should be zero (positions _poly_poisson-commute)
            for k in range(qi_qj.shape[0]):
                assert abs(qi_qj[k]) < 1e-10
            
            # Test {p_i, p_j} = 0
            pi_pj = _poly_poisson(pi, deg_q, pj, deg_q, PSI, CLMO, _ENCODE_DICT_GLOBAL)
            
            # Should be zero (momenta _poly_poisson-commute)
            for k in range(pi_pj.shape[0]):
                assert abs(pi_pj[k]) < 1e-10
            
            # Test {q_i, p_j} = delta _ij (Kronecker delta)
            qi_pj = _poly_poisson(qi, deg_q, pj, deg_q, PSI, CLMO, _ENCODE_DICT_GLOBAL)
            
            if i == j:
                # {q_i, p_i} should be 1
                # The polynomial should have a single constant term equal to 1
                assert qi_pj.shape[0] == 1  # Degree 0 polynomial (constant)
                assert abs(qi_pj[0] - 1.0) < 1e-10
            else:
                # {q_i, p_j} should be 0 for i ≠ j
                for k in range(qi_pj.shape[0]):
                    assert abs(qi_pj[k]) < 1e-10

def test_get_degree():
    """Test the _get_degree function for homogeneous polynomials"""
    # Test constant polynomial (degree 0)
    const_poly = _make_poly(0, PSI)
    assert _get_degree(const_poly, PSI) == 0

    # Test linear polynomial (degree 1)
    linear_poly = _make_poly(1, PSI)
    assert _get_degree(linear_poly, PSI) == 1

    # Test quadratic polynomial (degree 2)
    quad_poly = _make_poly(2, PSI)
    assert _get_degree(quad_poly, PSI) == 2

    # Test higher degree polynomial (degree TEST_MAX_DEG)
    high_deg_poly = _make_poly(TEST_MAX_DEG, PSI)
    assert _get_degree(high_deg_poly, PSI) == TEST_MAX_DEG
    
    # Test with a polynomial that has coefficients (values don't matter for _get_degree)
    poly_deg3 = _make_poly(3, PSI)
    k_x0_cubed = np.array([3, 0, 0, 0, 0, 0], dtype=np.int64)
    idx_x0_cubed = _encode_multiindex(k_x0_cubed, 3, _ENCODE_DICT_GLOBAL)
    if idx_x0_cubed != -1 and idx_x0_cubed < poly_deg3.shape[0]:
        poly_deg3[idx_x0_cubed] = 1.0 # Assign a dummy value
    assert _get_degree(poly_deg3, PSI) == 3

    # Test with a polynomial that might be all zeros but allocated for a certain degree
    all_zeros_deg4 = _make_poly(4, PSI) # this is already all zeros
    assert _get_degree(all_zeros_deg4, PSI) == 4


def test_poly_clean_inplace():
    """Test the _poly_clean_inplace function"""
    # Test with real polynomial
    poly = _make_poly(2, PSI)
    # Set some values
    for i in range(poly.shape[0]):
        if i % 3 == 0:
            poly[i] = 1.0  # Normal value
        elif i % 3 == 1:
            poly[i] = 1e-16  # Very small value (noise)
        else:
            poly[i] = 1e-8  # Small but significant value
    
    # Copy for later comparison
    poly_orig = poly.copy()
    
    # Clean with tolerance 1e-10 (should only remove the 1e-16 values)
    _poly_clean_inplace(poly, 1e-10)
    
    for i in range(poly.shape[0]):
        if i % 3 == 0:
            assert poly[i] == 1.0  # Normal values unchanged
        elif i % 3 == 1:
            assert poly[i] == 0.0  # Very small values zeroed
        else:
            assert poly[i] == 1e-8  # Small but significant values unchanged
    
    # Test with higher tolerance
    poly = poly_orig.copy()
    _poly_clean_inplace(poly, 1e-7)
    
    for i in range(poly.shape[0]):
        if i % 3 == 0:
            assert poly[i] == 1.0  # Normal values unchanged
        elif i % 3 == 1 or i % 3 == 2:
            assert poly[i] == 0.0  # All small values zeroed
    
    # Test with complex polynomial
    complex_poly = _make_poly(3, PSI)
    # Set some complex values
    for i in range(complex_poly.shape[0]):
        if i % 4 == 0:
            complex_poly[i] = complex(1.0, 1.0)  # Normal value
        elif i % 4 == 1:
            complex_poly[i] = complex(1e-16, 0.0)  # Small real part
        elif i % 4 == 2:
            complex_poly[i] = complex(0.0, 1e-16)  # Small imaginary part
        else:
            complex_poly[i] = complex(1e-16, 1e-16)  # Both parts small
    
    # Copy for comparison
    complex_orig = complex_poly.copy()
    
    # Clean with tolerance 1e-10
    _poly_clean_inplace(complex_poly, 1e-10)
    
    for i in range(complex_poly.shape[0]):
        if i % 4 == 0:
            assert complex_poly[i] == complex(1.0, 1.0)  # Normal values unchanged
        else:
            assert complex_poly[i] == 0.0  # All small values zeroed
    
    # Test with values exactly at tolerance threshold
    threshold_poly = _make_poly(1, PSI)
    tol = 1e-8
    threshold_poly[0] = tol  # Exactly at threshold
    threshold_poly[1] = tol + 1e-10  # Just above threshold
    threshold_poly[2] = tol - 1e-10  # Just below threshold
    
    _poly_clean_inplace(threshold_poly, tol)
    assert threshold_poly[0] == 0.0  # At threshold should be zeroed
    assert threshold_poly[1] != 0.0  # Just above should be kept
    assert threshold_poly[2] == 0.0  # Just below should be zeroed

def test_poly_clean():
    """Test the _poly_clean function (out-of-place cleaning)"""
    # Test with real polynomial
    poly = _make_poly(2, PSI)
    result = _make_poly(2, PSI)
    
    # Set some values
    for i in range(poly.shape[0]):
        if i % 3 == 0:
            poly[i] = 1.0  # Normal value
        elif i % 3 == 1:
            poly[i] = 1e-16  # Very small value (noise)
        else:
            poly[i] = 1e-8  # Small but significant value
    
    # Clean with tolerance 1e-10 (should only remove the 1e-16 values)
    _poly_clean(poly, 1e-10, result)
    
    # Verify original is unchanged
    for i in range(poly.shape[0]):
        if i % 3 == 0:
            assert poly[i] == 1.0
        elif i % 3 == 1:
            assert poly[i] == 1e-16
        else:
            assert poly[i] == 1e-8
    
    # Verify result is cleaned
    for i in range(result.shape[0]):
        if i % 3 == 0:
            assert result[i] == 1.0  # Normal values copied
        elif i % 3 == 1:
            assert result[i] == 0.0  # Very small values zeroed
        else:
            assert result[i] == 1e-8  # Small but significant values copied
    
    # Test with complex polynomial
    complex_poly = _make_poly(3, PSI)
    complex_result = _make_poly(3, PSI)
    
    # Set some complex values
    for i in range(complex_poly.shape[0]):
        if i % 4 == 0:
            complex_poly[i] = complex(1.0, 1.0)  # Normal value
        elif i % 4 == 1:
            complex_poly[i] = complex(1e-16, 0.0)  # Small real part
        elif i % 4 == 2:
            complex_poly[i] = complex(0.0, 1e-16)  # Small imaginary part
        else:
            complex_poly[i] = complex(1e-16, 1e-16)  # Both parts small
    
    # Clean with tolerance 1e-10
    _poly_clean(complex_poly, 1e-10, complex_result)
    
    # Verify original is unchanged
    for i in range(complex_poly.shape[0]):
        if i % 4 == 0:
            assert complex_poly[i] == complex(1.0, 1.0)
        elif i % 4 == 1:
            assert complex_poly[i] == complex(1e-16, 0.0)
        elif i % 4 == 2:
            assert complex_poly[i] == complex(0.0, 1e-16)
        else:
            assert complex_poly[i] == complex(1e-16, 1e-16)
    
    # Verify result is cleaned
    for i in range(complex_result.shape[0]):
        if i % 4 == 0:
            assert complex_result[i] == complex(1.0, 1.0)  # Normal values copied
        else:
            assert complex_result[i] == 0.0  # All small values zeroed
    
    # Test with zero tolerance (should not modify anything)
    zero_tol_result = _make_poly(2, PSI)
    _poly_clean(poly, 0.0, zero_tol_result)
    
    # All values should be copied exactly
    for i in range(poly.shape[0]):
        assert zero_tol_result[i] == poly[i]

@pytest.mark.parametrize("degree", range(TEST_MAX_DEG + 1))
def test_poly_evaluate_zero_polynomial(degree):
    """Test evaluation of a zero homogeneous polynomial."""
    coeffs = _make_poly(degree, PSI) # all zeros
    point_vals = np.random.rand(N_VARS) + 1j * np.random.rand(N_VARS)
    
    numeric_eval = _poly_evaluate(coeffs, degree, point_vals, CLMO)
    assert np.isclose(numeric_eval, 0.0 + 0.0j)

@pytest.mark.parametrize("degree", range(1, TEST_MAX_DEG + 1)) # Skip degree 0 for monomial test
def test_poly_evaluate_single_monomial(degree):
    """Test evaluation of a single monomial."""
    coeffs = _make_poly(degree, PSI)
    
    # Create a single monomial, e.g., 2.5 * x_0^degree (if degree > 0)
    # or 2.5 * x_0^{d_0} * x_1^{d_1} ... such that sum(d_i) = degree
    # For simplicity, let's use x_0^degree if possible, or the first available monomial.
    
    test_coeff_val = 2.5 - 1.5j
    
    # Create a specific monomial k_test (e.g., x0^d)
    k_test = np.zeros(N_VARS, dtype=np.int64)
    # Distribute degree among first few vars if possible
    vars_to_use = min(N_VARS, degree)
    if vars_to_use > 0:
        deg_per_var = degree // vars_to_use
        remainder = degree % vars_to_use
        for i in range(vars_to_use):
            k_test[i] = deg_per_var
            if i < remainder:
                k_test[i] +=1
    else: # degree is 0, this case is not hit by parametrize
        pass

    idx = _encode_multiindex(k_test, degree, _ENCODE_DICT_GLOBAL)
    if idx != -1 and idx < coeffs.shape[0]:
        coeffs[idx] = test_coeff_val
    else: # Fallback if the specific k_test is not found (should not happen for valid k)
        if coeffs.shape[0] > 0:
            coeffs[0] = test_coeff_val # Assign to the first available coefficient slot
            k_test = _decode_multiindex(0, degree, CLMO) # And use its exponents
        else: # No coefficients for this degree (e.g. PSI[N_VARS, degree] is 0)
             pytest.skip(f"No monomials for degree {degree} with N_VARS={N_VARS}")

    point_vals_np = np.random.rand(N_VARS) * 2 - 1 + 1j * (np.random.rand(N_VARS) * 2 - 1) # Random values in [-1,1] + j*[-1,1]
    
    numeric_eval = _poly_evaluate(coeffs, degree, point_vals_np, CLMO)
    
    # Compute expected value manually: test_coeff_val * product(point_vals_np[i]^k_test[i])
    expected_val = test_coeff_val
    for i in range(N_VARS):
        if k_test[i] > 0:
            expected_val *= point_vals_np[i] ** k_test[i]
    
    assert np.isclose(numeric_eval, expected_val)

@pytest.mark.parametrize("degree", range(TEST_MAX_DEG + 1))
def test_poly_evaluate_multiple_terms(degree):
    coeffs = _make_poly(degree, PSI)
    
    # Assign some random complex coefficients
    num_coeffs_to_set = min(coeffs.shape[0], 5) # Set up to 5 random coeffs
    if num_coeffs_to_set == 0:
        pytest.skip(f"No coefficients available for degree {degree}")
        
    indices_to_set = np.random.choice(coeffs.shape[0], num_coeffs_to_set, replace=False)
    
    for i in indices_to_set:
        coeffs[i] = np.random.rand() - 0.5 + 1j*(np.random.rand() - 0.5)

    point_vals_np = np.random.rand(N_VARS) + 1j*np.random.rand(N_VARS)
    
    numeric_eval = _poly_evaluate(coeffs, degree, point_vals_np, CLMO)
    
    # Compute expected value manually
    expected_val = 0.0 + 0.0j
    for i in indices_to_set:
        k = _decode_multiindex(i, degree, CLMO)
        term_val = coeffs[i]
        for j in range(N_VARS):
            if k[j] > 0:
                term_val *= point_vals_np[j] ** k[j]
        expected_val += term_val
    
    assert np.isclose(numeric_eval, expected_val)

def test_poly_evaluate_at_origin():
    """Test evaluation at the origin point."""
    for degree in range(TEST_MAX_DEG + 1):
        coeffs = _make_poly(degree, PSI)
        if coeffs.shape[0] > 0:
            coeffs[np.random.randint(0, coeffs.shape[0])] = 1.0 + 1.0j # Make it non-zero

        point_at_origin = np.zeros(N_VARS, dtype=np.complex128)
        numeric_eval = _poly_evaluate(coeffs, degree, point_at_origin, CLMO)

        if degree == 0:
            if coeffs.shape[0] > 0:
                assert np.isclose(numeric_eval, coeffs[0])
            else: # Should not happen if PSI[N_VARS,0] is 1
                assert np.isclose(numeric_eval, 0.0 + 0.0j)
        else:
            assert np.isclose(numeric_eval, 0.0 + 0.0j)

def test_poly_evaluate_point_with_zeros():
    """Test evaluation at a point with some zero components."""
    degree = 2
    if PSI[N_VARS, degree] == 0: pytest.skip("Not enough terms for degree 2 test")

    coeffs = _make_poly(degree, PSI)
    # Example: P = x0^2 + x0*x1 + x1^2
    k_x0sq = np.array([2,0,0,0,0,0], dtype=np.int64)
    k_x0x1 = np.array([1,1,0,0,0,0], dtype=np.int64)
    k_x1sq = np.array([0,2,0,0,0,0], dtype=np.int64)

    idx_x0sq = _encode_multiindex(k_x0sq, degree, _ENCODE_DICT_GLOBAL)
    idx_x0x1 = _encode_multiindex(k_x0x1, degree, _ENCODE_DICT_GLOBAL)
    idx_x1sq = _encode_multiindex(k_x1sq, degree, _ENCODE_DICT_GLOBAL)

    if idx_x0sq != -1: coeffs[idx_x0sq] = 1.0
    if idx_x0x1 != -1: coeffs[idx_x0x1] = 1.0
    if idx_x1sq != -1: coeffs[idx_x1sq] = 1.0
    
    point_vals_np = np.zeros(N_VARS, dtype=np.complex128)
    point_vals_np[0] = 2.0 + 1j # x0 = 2+j
    # x1 and others are 0
    
    # Expected: (2+j)^2 + (2+j)*0 + 0^2 = (2+j)^2 = 4 + 4j - 1 = 3 + 4j
    expected_eval = (2.0+1j)**2

    numeric_eval = _poly_evaluate(coeffs, degree, point_vals_np, CLMO)
    
    assert np.isclose(numeric_eval, expected_eval)



"""Example script: Detecting heteroclinic connections between two manifolds.

Run with
    python examples/heteroclinic_connection.py
"""

import os
import sys

sys.path.append(os.path.join(os.path.dirname(__file__), "..", "src"))

from hiten.algorithms.connections import Connection, SearchConfig
from hiten.algorithms.poincare import SynodicMapConfig
from hiten.system import System


def main() -> None:
    system = System.from_bodies("earth", "moon")
    mu = system.mu

    l1 = system.get_libration_point(1)
    l2 = system.get_libration_point(2)

    halo_l1 = l1.create_orbit('halo', amplitude_z=0.5, zenith='southern')
    halo_l1.correct()
    halo_l1.propagate()

    halo_l2 = l2.create_orbit('halo', amplitude_z=0.3663368, zenith='northern')
    halo_l2.correct()
    halo_l2.propagate()

    manifold_l1 = halo_l1.manifold(stable=True, direction='positive')
    manifold_l1.compute(integration_fraction=0.9, step=0.005)

    manifold_l2 = halo_l2.manifold(stable=False, direction='negative')
    manifold_l2.compute(integration_fraction=1.0, step=0.005)

    section_cfg = SynodicMapConfig(
        section_axis="x",
        section_offset=1 - mu,
        plane_coords=("y", "z"),
        interp_kind="cubic",
        segment_refine=30,
        tol_on_surface=1e-9,
        dedup_time_tol=1e-9,
        dedup_point_tol=1e-9,
        max_hits_per_traj=None,
        n_workers=None,
    )

    conn = Connection(
        section=section_cfg,
        direction=None,
        search_cfg=SearchConfig(delta_v_tol=1, ballistic_tol=1e-8, eps2d=1e-3),
    )

    conn.solve(manifold_l1, manifold_l2)
    print(conn)
    conn.plot(dark_mode=True)

if __name__ == "__main__":
    main()


"""Example script: Generating a Poincaré map for the synodic section of a vertical orbit manifold.

Run with
    python examples/heteroclinic_connection.py
"""

import os
import sys

sys.path.append(os.path.join(os.path.dirname(__file__), "..", "src"))

from hiten.algorithms.poincare import SynodicMap, SynodicMapConfig
from hiten.system import System, VerticalOrbit


def main() -> None:
    system = System.from_bodies("earth", "moon")
    l_point = system.get_libration_point(1)

    cm = l_point.get_center_manifold(degree=6)
    cm.compute()

    ic_seed = cm.ic([0.0, 0.0], 0.6, "q3") # Good initial guess from CM

    orbit = VerticalOrbit(l_point, initial_state=ic_seed)
    orbit.correct(max_attempts=100, finite_difference=True)
    orbit.propagate(steps=1000)

    manifold = orbit.manifold(stable=True, direction="positive")
    manifold.compute(step=0.005)
    manifold.plot()

    section_cfg = SynodicMapConfig(
        section_axis="y",
        section_offset=0.0,
        plane_coords=("x", "z"),
        interp_kind="cubic",
        segment_refine=30,
        newton_max_iter=10,
    )
    synodic_map = SynodicMap(section_cfg)
    synodic_map.from_manifold(manifold)
    synodic_map.plot()


if __name__ == "__main__":
    main()


r"""
hiten.algorithms.integrators.rk
=========================

Explicit Runge-Kutta integrators used throughout the project.  Both fixed
and adaptive step-size variants are provided together with small convenience
factories that select an appropriate implementation given the desired
formal order of accuracy.

The concrete schemes implemented are:

* Fixed step: _RK4, _RK6, _RK8
* Adaptive embedded: _RK45 (Dormand-Prince) and _DOP853 (Dormand-Prince 8(5,3))

Internally the module also defines helper routines to evaluate Hamiltonian
vector fields with :pyfunc:`numba.njit` and to wrap right-hand side (RHS)
callables into a uniform signature accepted by the integrators.

References
----------
Hairer, E.; Nørsett, S.; Wanner, G. (1993). "Solving Ordinary Differential
Equations I".

Dormand, J. R.; Prince, P. J. (1980). "A family of embedded Runge-Kutta
formulas".
"""

import inspect
from typing import Callable, Optional

import numpy as np

from hiten.algorithms.dynamics.base import _DynamicalSystem
from hiten.algorithms.integrators.base import _Integrator, _Solution
from hiten.algorithms.integrators.coefficients.dop853 import E3 as DOP853_E3
from hiten.algorithms.integrators.coefficients.dop853 import E5 as DOP853_E5
from hiten.algorithms.integrators.coefficients.dop853 import \
    INTERPOLATOR_POWER as DOP853_INTERPOLATOR_POWER
from hiten.algorithms.integrators.coefficients.dop853 import \
    N_STAGES as DOP853_N_STAGES
from hiten.algorithms.integrators.coefficients.dop853 import \
    N_STAGES_EXTENDED as DOP853_N_STAGES_EXTENDED
from hiten.algorithms.integrators.coefficients.dop853 import A as DOP853_A
from hiten.algorithms.integrators.coefficients.dop853 import B as DOP853_B
from hiten.algorithms.integrators.coefficients.dop853 import C as DOP853_C
from hiten.algorithms.integrators.coefficients.dop853 import D as DOP853_D
from hiten.algorithms.integrators.coefficients.rk4 import A as RK4_A
from hiten.algorithms.integrators.coefficients.rk4 import B as RK4_B
from hiten.algorithms.integrators.coefficients.rk4 import C as RK4_C
from hiten.algorithms.integrators.coefficients.rk6 import A as RK6_A
from hiten.algorithms.integrators.coefficients.rk6 import B as RK6_B
from hiten.algorithms.integrators.coefficients.rk6 import C as RK6_C
from hiten.algorithms.integrators.coefficients.rk8 import A as RK8_A
from hiten.algorithms.integrators.coefficients.rk8 import B as RK8_B
from hiten.algorithms.integrators.coefficients.rk8 import C as RK8_C
from hiten.algorithms.integrators.coefficients.rk45 import \
    B_HIGH as RK45_B_HIGH
from hiten.algorithms.integrators.coefficients.rk45 import B_LOW as RK45_B_LOW
from hiten.algorithms.integrators.coefficients.rk45 import A as RK45_A
from hiten.algorithms.integrators.coefficients.rk45 import C as RK45_C
from hiten.algorithms.integrators.coefficients.rk45 import E as RK45_E
from hiten.algorithms.utils.config import TOL
from hiten.utils.log_config import logger


class _RungeKuttaBase(_Integrator):
    r"""
    Shared functionality of explicit Runge-Kutta schemes.

    The class stores a Butcher tableau and provides a single low level helper
    :pyfunc:`_rk_embedded_step` that advances one macro time step and, when a
    second set of weights is available, returns an error estimate suitable
    for adaptive step-size control.

    Attributes
    ----------
    _A : numpy.ndarray of shape (s, s)
        Strictly lower triangular array of stage coefficients :math:`a_{ij}`.
    _B_HIGH : numpy.ndarray of shape (s,)
        Weights of the high order solution.
    _B_LOW : numpy.ndarray or None
        Weights of the lower order solution, optional.  When *None* no error
        estimate is produced and :pyfunc:`_rk_embedded_step` falls back to
        the high order result for both outputs.
    _C : numpy.ndarray of shape (s,)
        Nodes :math:`c_i` measured in units of the step size.
    _p : int
        Formal order of accuracy of the high order scheme.

    Notes
    -----
    The class is **not** intended to be used directly.  Concrete subclasses
    define the specific coefficients and expose a public interface compliant
    with :class:`hiten.algorithms.integrators.base._Integrator`.
    """

    _A: np.ndarray = None
    _B_HIGH: np.ndarray = None
    _B_LOW: Optional[np.ndarray] = None
    _C: np.ndarray = None
    _p: int = 0

    def _rk_embedded_step(self, f, t, y, h):
        s = self._B_HIGH.size
        k = np.empty((s, y.size), dtype=np.float64)

        k[0] = f(t, y)
        for i in range(1, s):
            y_stage = y.copy()
            for j in range(i):
                a_ij = self._A[i, j]
                if a_ij != 0.0:
                    y_stage += h * a_ij * k[j]
            k[i] = f(t + self._C[i] * h, y_stage)

        y_high = y + h * np.dot(self._B_HIGH, k)

        if self._B_LOW is not None:
            y_low = y + h * np.dot(self._B_LOW, k)
        else:
            y_low = y_high.copy()
        err_vec = y_high - y_low
        return y_high, y_low, err_vec


class _FixedStepRK(_RungeKuttaBase):
    r"""Explicit fixed-step Runge-Kutta scheme.

    Parameters
    ----------
    name : str
        Human readable identifier of the scheme (e.g. ``"_RK4"``).
    A, B, C : numpy.ndarray
        Butcher tableau as returned by :pymod:`hiten.algorithms.integrators.coefficients.*`.
    order : int
        Formal order of accuracy :math:`p` of the method.
    **options
        Additional keyword options forwarded to the base :class:`_Integrator`.

    Notes
    -----
    The step size is assumed to be **constant** and is inferred from the
    spacing of the *t_vals* array supplied to :pyfunc:`integrate`.
    """

    def __init__(self, name: str, A: np.ndarray, B: np.ndarray, C: np.ndarray, order: int, **options):
        self._A = A
        self._B_HIGH = B
        self._B_LOW = None
        self._C = C
        self._p = order
        super().__init__(name, **options)

    @property
    def order(self) -> int:
        return self._p

    def integrate(
        self,
        system: _DynamicalSystem,
        y0: np.ndarray,
        t_vals: np.ndarray,
        **kwargs,
    ) -> _Solution:
        self.validate_inputs(system, y0, t_vals)

        rhs_wrapped = _build_rhs_wrapper(system)

        # The RHS that comes in may already embed the intended time direction.
        # Therefore do *not* apply any additional sign here - simply forward
        # the call to the wrapped system RHS.
        def f(t, y):
            return rhs_wrapped(t, y)

        traj = np.empty((t_vals.size, y0.size), dtype=np.float64)
        derivs = np.empty_like(traj)

        # Initial state and derivative
        traj[0] = y0.copy()
        derivs[0] = f(t_vals[0], y0)

        for idx in range(t_vals.size - 1):
            t_n = t_vals[idx]
            h = t_vals[idx + 1] - t_n
            y_n = traj[idx]

            # Perform RK step and obtain high-order solution
            y_high, _, _ = self._rk_embedded_step(f, t_n, y_n, h)
            traj[idx + 1] = y_high

            # Derivative at the new time point (needed for Hermite interpolation)
            derivs[idx + 1] = f(t_vals[idx + 1], y_high)

        return _Solution(times=t_vals.copy(), states=traj, derivatives=derivs)


class _RK4(_FixedStepRK):
    def __init__(self, **opts):
        super().__init__("_RK4", RK4_A, RK4_B, RK4_C, 4, **opts)


class _RK6(_FixedStepRK):
    def __init__(self, **opts):
        super().__init__("_RK6", RK6_A, RK6_B, RK6_C, 6, **opts)


class _RK8(_FixedStepRK):
    def __init__(self, **opts):
        super().__init__("_RK8", RK8_A, RK8_B, RK8_C, 8, **opts)


class _AdaptiveStepRK(_RungeKuttaBase):
    r"""
    Embedded adaptive Runge-Kutta integrator with PI controller.

    The class implements proportional-integral (PI) step-size control using
    the error estimates returned by :pyfunc:`_rk_embedded_step`.  Two safety
    factors are used:

    * *rtol*, *atol*  - user requested relative and absolute tolerances
    * :pyattr:`SAFETY` - hard coded damping on the acceptance criterion.

    Parameters
    ----------
    name : str, default "AdaptiveRK"
        Identifier passed to the :class:`_Integrator` base class.
    rtol, atol : float, optional
        Relative and absolute error tolerances.  Defaults are read from
        :pydata:`hiten.utils.config.TOL`.
    max_step : float, optional
        Upper bound on the step size.  :math:`\infty` by default.
    min_step : float or None, optional
        Lower bound on the step size.  When *None* the value is derived from
        machine precision.

    Attributes
    ----------
    SAFETY, MIN_FACTOR, MAX_FACTOR : float
        Magic constants used by the PI controller.  They follow SciPy's
        implementation and the recommendations by Hairer et al.

    Raises
    ------
    RuntimeError
        If the step size underflows while trying to satisfy the error
        tolerance.
    """

    SAFETY = 0.9
    MIN_FACTOR = 0.2
    MAX_FACTOR = 10.0

    def __init__(self,
                 name: str = "AdaptiveRK",
                 rtol: float = TOL,
                 atol: float = TOL,
                 max_step: float = np.inf,
                 min_step: Optional[float] = None,
                 **options):
        super().__init__(name, **options)
        self._rtol = rtol
        self._atol = atol
        self._max_step = max_step
        if min_step is None:
            self._min_step = 10.0 * np.finfo(float).eps
        else:
            self._min_step = min_step
        if not hasattr(self, "_err_exp") or self._err_exp == 0:
            self._err_exp = 1.0 / (self._p)

    @property
    def order(self) -> int:
        return self._p

    def integrate(self, system: _DynamicalSystem, y0: np.ndarray, t_vals: np.ndarray, **kwargs) -> _Solution:
        r"""
        Integrate a dynamical system using an adaptive Runge-Kutta method.
        """
        self.validate_inputs(system, y0, t_vals)

        debug = self.options.get("debug", False)

        rhs_wrapped = _build_rhs_wrapper(system)

        # The RHS that comes in may already embed the intended time direction.
        # Therefore do *not* apply any additional sign here - simply forward
        # the call to the wrapped system RHS.
        def f(t, y):
            return rhs_wrapped(t, y)

        t_span = (t_vals[0], t_vals[-1])

        t = t_span[0]
        y = np.ascontiguousarray(y0, dtype=np.float64)
        ts, ys, dys = [t], [y.copy()], [f(t, y)]

        h = self._select_initial_step(f, t, y, t_span[1])
        h = max(h, 1e-4 * abs(t_span[1] - t_span[0]))

        err_prev = None
        accepted_steps = 0
        step_rejected = False

        while (t - t_span[1]) * 1 < 0:
            min_step = 10 * abs(np.nextafter(t, t + 1) - t)
            if h > self._max_step:
                h = self._max_step
            if t + 1 * h > t_span[1]:
                h = abs(t_span[1] - t)

            y_high, y_low, err_vec = self._rk_embedded_step(f, t, y, h)
            scale = self._atol + self._rtol * np.maximum(np.abs(y), np.abs(y_high))
            err_norm = np.linalg.norm(err_vec / scale) / np.sqrt(err_vec.size)

            if err_norm <= 1.0:
                t_new = t + 1 * h
                y_new = y_high

                ts.append(t_new)
                ys.append(y_new.copy())
                dys.append(f(t_new, y_new))
                t, y = t_new, y_new

                beta = 1.0 / (self._p + 1)
                alpha = 0.4 * beta
                if err_prev is None:
                    factor = self.SAFETY * err_norm ** (-beta)
                else:
                    factor = self.SAFETY * err_norm ** (-beta) * err_prev ** alpha
                factor = np.clip(factor, self.MIN_FACTOR, self.MAX_FACTOR)
                if step_rejected:
                    factor = min(1.0, factor)

                h *= factor
                step_rejected = False
                err_prev = err_norm

                accepted_steps += 1
                if debug and accepted_steps % 1000 == 0:
                    logger.info(f"[AdaptiveRK] step {accepted_steps:7d}: t={t:.6g} h={h:.3g} err={err_norm:.3g}")
            else:
                factor = max(self.MIN_FACTOR, self.SAFETY * err_norm ** (-self._err_exp))
                h *= factor
                step_rejected = True
                if abs(h) < min_step:
                    raise RuntimeError("Step size underflow in adaptive RK integrator.")

        # Convert lists to arrays for vectorized interpolation
        ts = np.array(ts)
        ys = np.array(ys)
        dys = np.array(dys)
        t_eval = np.asarray(t_vals)

        # Vectorized Hermite interpolation for all t_eval
        indices = np.searchsorted(ts, t_eval, side='right') - 1
        indices = np.clip(indices, 0, len(ts) - 2)

        t0s = ts[indices]
        t1s = ts[indices + 1]
        y0s = ys[indices]
        y1s = ys[indices + 1]
        dy0s = dys[indices]
        dy1s = dys[indices + 1]
        taus = (t_eval - t0s) / (t1s - t0s)

        tau2 = taus ** 2
        tau3 = tau2 * taus
        h = t1s - t0s
        h00 = 2 * tau3 - 3 * tau2 + 1
        h10 = tau3 - 2 * tau2 + taus
        h01 = -2 * tau3 + 3 * tau2
        h11 = tau3 - tau2

        y_eval = (
            h00[:, None] * y0s +
            h10[:, None] * h[:, None] * dy0s +
            h01[:, None] * y1s +
            h11[:, None] * h[:, None] * dy1s
        )

        # Also compute derivatives at t_eval if needed
        # For now, use f(t_eval, y_eval) for derivatives
        derivs_out = np.array([f(t, y) for t, y in zip(t_eval, y_eval)])

        return _Solution(times=t_eval.copy(), states=y_eval, derivatives=derivs_out)

    def _select_initial_step(self, f, t0, y0, tf):
        r"""
        Choose an initial step size following Hairer et al. / SciPy heuristics.

        The strategy tries to obtain a first step that keeps the truncation
        error around the requested tolerance.  A too-small *h* makes the
        adaptive solver crawl, while a too-large *h* triggers many rejected
        steps.  This implementation closely mirrors SciPy's `_initial_step`.
        """
        # Compute derivative at the initial point.
        dy0 = f(t0, y0)

        # Weighted norms used by the error control.
        scale = self._atol + self._rtol * np.abs(y0)
        d0 = np.linalg.norm(y0 / scale) / np.sqrt(y0.size)
        d1 = np.linalg.norm(dy0 / scale) / np.sqrt(y0.size)

        if d0 < 1e-5 or d1 < 1e-5:
            h0 = 1e-6
        else:
            h0 = 0.01 * d0 / d1

        # Try a tentative first step and look at the curvature.
        y1 = y0 + h0 * dy0
        dy1 = f(t0 + h0, y1)

        d2 = np.linalg.norm((dy1 - dy0) / scale) / np.sqrt(y0.size) / h0

        if max(d1, d2) <= 1e-15:
            h1 = max(1e-6, 0.1 * abs(tf - t0))
        else:
            h1 = (0.01 / max(d1, d2)) ** (1.0 / (self._p + 1))

        h = min(100 * h0, h1)

        # Respect user-supplied step limits.
        h = min(h, abs(tf - t0), self._max_step)
        h = max(h, self._min_step)

        return h

    def _update_factor(self, err_norm):
        return np.clip(self.SAFETY * err_norm ** (-self._err_exp), self.MIN_FACTOR, self.MAX_FACTOR)


class _RK45(_AdaptiveStepRK):
    _A = RK45_A
    _B_HIGH = RK45_B_HIGH
    _B_LOW = None
    _C = RK45_C
    _p = 5
    _err_exp = 1.0 / 5.0
    _E = RK45_E

    def __init__(self, **opts):
        super().__init__("_RK45", **opts)

    def _rk_embedded_step(self, f, t, y, h):
        s = 6
        k = np.empty((s + 1, y.size))
        k[0] = f(t, y)
        for i in range(1, s):
            y_stage = y + h * (self._A[i, :i] @ k[:i])
            k[i] = f(t + self._C[i] * h, y_stage)
        y_high = y + h * (self._B_HIGH @ k[:s])
        k[s] = f(t + h, y_high)
        err_vec = h * (k.T @ self._E)
        y_low = y_high - err_vec
        return y_high, y_low, err_vec


class _DOP853(_AdaptiveStepRK):
    _A = DOP853_A[:DOP853_N_STAGES, :DOP853_N_STAGES]
    _B_HIGH = DOP853_B[:DOP853_N_STAGES]
    _B_LOW = None
    _C = DOP853_C[:DOP853_N_STAGES]

    _p = 8
    _err_exp = 1.0 / _p

    _E3 = DOP853_E3
    _E5 = DOP853_E5
    _N_STAGES = DOP853_N_STAGES

    def __init__(self, **opts):
        super().__init__("_DOP853", **opts)

    def _rk_embedded_step(self, f, t, y, h):
        s = self._N_STAGES
        k = np.empty((s + 1, y.size), dtype=np.float64)
        k[0] = f(t, y)
        for i in range(1, s):
            y_stage = y.copy()
            for j in range(i):
                a_ij = self._A[i, j]
                if a_ij != 0.0:
                    y_stage += h * a_ij * k[j]
            k[i] = f(t + self._C[i] * h, y_stage)
        y_high = y.copy()
        for j in range(s):
            b_j = self._B_HIGH[j]
            if b_j != 0.0:
                y_high += h * b_j * k[j]
        k[s] = f(t + h, y_high)
        err_vec = self._estimate_error(k, h)
        y_low = y_high - err_vec
        return y_high, y_low, err_vec

    def _estimate_error(self, K, h):
        err5 = np.dot(K.T, self._E5)
        err3 = np.dot(K.T, self._E3)
        denom = np.hypot(np.abs(err5), 0.1 * np.abs(err3))
        correction_factor = np.ones_like(err5)
        mask = denom > 0
        correction_factor[mask] = np.abs(err5[mask]) / denom[mask]
        return h * err5 * correction_factor

    def _estimate_error_norm(self, K, h, scale):
        err5 = np.dot(K.T, self._E5) / scale
        err3 = np.dot(K.T, self._E3) / scale
        err5_norm_2 = np.linalg.norm(err5) ** 2
        err3_norm_2 = np.linalg.norm(err3) ** 2
        if err5_norm_2 == 0 and err3_norm_2 == 0:
            return 0.0
        denom = err5_norm_2 + 0.01 * err3_norm_2
        return np.abs(h) * err5_norm_2 / np.sqrt(denom * len(scale))


class RungeKutta:
    _map = {4: _RK4, 6: _RK6, 8: _RK8}
    def __new__(cls, order=4, **opts):
        if order not in cls._map:
            raise ValueError("RK order must be 4, 6, or 8")
        return cls._map[order](**opts)

class AdaptiveRK:
    _map = {5: _RK45, 8: _DOP853}
    def __new__(cls, order=5, **opts):
        if order not in cls._map:
            raise ValueError("Adaptive RK order not supported")
        return cls._map[order](**opts)


def _build_rhs_wrapper(system: _DynamicalSystem) -> Callable[[float, np.ndarray], np.ndarray]:
    r"""
    Return a JIT friendly wrapper around :pyfunc:`hiten.system.rhs`.

    The dynamical systems implemented in the code base expose their vector
    field either as ``rhs(t, y)`` or, for autonomous systems, as ``rhs(y)``.
    The integrator layer expects the non autonomous signature and therefore
    needs to adapt the call site on the fly. This helper inspects the right
    hand side via :pyfunc:`inspect.signature` and generates a small
    :pyfunc:`numba.njit` compiled closure with the correct arity.

    Parameters
    ----------
    system : _DynamicalSystem
        Instance providing the original vector field.

    Returns
    -------
    Callable[[float, numpy.ndarray], numpy.ndarray]
        Function accepting the full ``(t, y)`` signature required by the
        integrators.

    Raises
    ------
    ValueError
        If the detected signature contains neither one nor two positional
        arguments.
    """

    rhs_func = system.rhs

    sig = inspect.signature(rhs_func)
    n_params = len(sig.parameters)

    # Case 1: Function already accepts (t, y) - use directly.
    if n_params >= 2:
        return rhs_func

    # Case 2: Autonomous system with signature rhs(y).  Wrap to inject time argument.
    if n_params == 1:

        def _rhs_one(t, y):
            return rhs_func(y)

        return _rhs_one

    raise ValueError(
        f"Unsupported rhs signature with {n_params} parameters. "
        "Only (t, y) or (y,) are currently supported."
    )



"""
Public API for the ``system`` package.

This module re-exports the most frequently used classes so that users can
simply write::

    from system import System, Body, L1Point, HaloOrbit

instead of navigating the full internal hierarchy (``hiten.system.base``,
``hiten.system.libration.collinear`` ...).
"""

from ..algorithms.utils.types import SynodicState
from ..algorithms.poincare.centermanifold.base import CenterManifoldMap
from ..algorithms.poincare.centermanifold.config import _CenterManifoldMapConfig
from .base import System
# Core containers
from .body import Body
# Center manifold
from .center import CenterManifold
from .family import OrbitFamily
# Libration points
from .libration.base import LibrationPoint, LinearData
from .libration.collinear import CollinearPoint, L1Point, L2Point, L3Point
from .libration.triangular import L4Point, L5Point, TriangularPoint
from .manifold import Manifold, ManifoldResult
# Orbits
from .orbits.base import GenericOrbit, PeriodicOrbit
from .orbits.halo import HaloOrbit
from .orbits.lyapunov import LyapunovOrbit
from .orbits.vertical import VerticalOrbit

__all__ = [
    # Base system
    "Body",
    "System",
    "ManifoldResult",
    "Manifold",
    # Libration points
    "LinearData",
    "LibrationPoint",
    "CollinearPoint",
    "TriangularPoint",
    "L1Point",
    "L2Point",
    "L3Point",
    "L4Point",
    "L5Point",
    # Center manifold
    "CenterManifold",
    # Poincare map
    "_CenterManifoldMapConfig",
    "CenterManifoldMap",
    # Orbits / configs
    "PeriodicOrbit",
    "GenericOrbit",
    "HaloOrbit",
    "LyapunovOrbit",
    "VerticalOrbit",
    "SynodicState",
    # Family
    "OrbitFamily",
]



import numpy as np
import pytest
import sympy as sp
from numba.typed import List

from hiten.algorithms.polynomial.base import (_CLMO_GLOBAL, _ENCODE_DICT_GLOBAL,
                                               _PSI_GLOBAL, _encode_multiindex,
                                               _make_poly)
from hiten.algorithms.polynomial.conversion import poly2sympy, sympy2poly
from hiten.algorithms.utils.config import N_VARS

s_vars = list(sp.symbols(f'x_0:{N_VARS}'))
TEST_MAX_DEG = _PSI_GLOBAL.shape[1] - 1


def _create_custom_poly_list(max_deg: int, terms: dict) -> List[np.ndarray]:
    py_list_of_coeffs = [_make_poly(d, _PSI_GLOBAL) for d in range(max_deg + 1)]

    for k_tuple, coeff_val in terms.items():
        k_np = np.array(k_tuple, dtype=np.int64)
        deg = int(sum(k_np))
        pos = _encode_multiindex(k_np, deg, _ENCODE_DICT_GLOBAL)
        py_list_of_coeffs[deg][pos] = complex(coeff_val)
    
    numba_list = List()
    for arr in py_list_of_coeffs:
        numba_list.append(arr)
    return numba_list


def _compare_poly_lists(list1: List[np.ndarray], list2: List[np.ndarray], tol=1e-12) -> bool:
    if len(list1) != len(list2):
        return False

    for i in range(len(list1)):
        arr1, arr2 = list1[i], list2[i]
        if not np.allclose(arr1, arr2, atol=tol, rtol=tol):
            print(f"Data mismatch at degree {i}:\nArr1: {arr1}\nArr2: {arr2}")
            return False
    return True


def test_poly2sympy_zero():
    poly_list_zero = _create_custom_poly_list(0, {}) # Degree 0, no terms
    expr = poly2sympy(poly_list_zero, s_vars, _PSI_GLOBAL, _CLMO_GLOBAL)
    assert expr == sp.Integer(0)

    poly_list_zero_deg2 = _create_custom_poly_list(2, {}) # Up to degree 2, all zero
    expr2 = poly2sympy(poly_list_zero_deg2, s_vars, _PSI_GLOBAL, _CLMO_GLOBAL)
    assert expr2 == sp.Integer(0)
    
    empty_numba_list = List() 
    expr_empty = poly2sympy(empty_numba_list, s_vars, _PSI_GLOBAL, _CLMO_GLOBAL)
    assert expr_empty == sp.Integer(0)


def test_sympy2poly_zero():
    expr = sp.Integer(0)
    poly_list = sympy2poly(expr, s_vars, _PSI_GLOBAL, _CLMO_GLOBAL, _ENCODE_DICT_GLOBAL)
    
    expected_coeffs = _make_poly(0, _PSI_GLOBAL)
    expected_list = List()
    expected_list.append(expected_coeffs)
    
    assert _compare_poly_lists(poly_list, expected_list)


def test_poly2sympy_constant():
    const_val = 5.5
    poly_list = _create_custom_poly_list(0, {(0,)*N_VARS: const_val})
    expr = poly2sympy(poly_list, s_vars, _PSI_GLOBAL, _CLMO_GLOBAL)
    assert expr == sp.Float(const_val)


def test_sympy2poly_constant():
    const_val = 7.0
    expr = sp.Float(const_val)
    poly_list = sympy2poly(expr, s_vars, _PSI_GLOBAL, _CLMO_GLOBAL, _ENCODE_DICT_GLOBAL)
    
    expected_coeffs = _make_poly(0, _PSI_GLOBAL)
    expected_coeffs[_encode_multiindex(np.zeros(N_VARS, dtype=np.int64), 0, _ENCODE_DICT_GLOBAL)] = const_val
    expected_list = List()
    expected_list.append(expected_coeffs)
    
    assert _compare_poly_lists(poly_list, expected_list)


def test_poly2sympy_single_variable():
    target_var_idx = 1
    k_tuple = [0]*N_VARS
    k_tuple[target_var_idx] = 1
    k_tuple = tuple(k_tuple)

    poly_list = _create_custom_poly_list(1, {k_tuple: 1.0})
    expr = poly2sympy(poly_list, s_vars, _PSI_GLOBAL, _CLMO_GLOBAL)
    assert sp.simplify(expr - s_vars[target_var_idx]) == 0


def test_sympy2poly_single_variable():
    target_var_idx = 2
    expr = s_vars[target_var_idx]
    poly_list = sympy2poly(expr, s_vars, _PSI_GLOBAL, _CLMO_GLOBAL, _ENCODE_DICT_GLOBAL)

    k_tuple = [0]*N_VARS
    k_tuple[target_var_idx] = 1
    k_tuple = tuple(k_tuple)
    
    expected_list = _create_custom_poly_list(1, {k_tuple: 1.0})
    assert _compare_poly_lists(poly_list, expected_list)


def test_round_trip():
    # P(x) = 2.0*x_0 + 3.0*x_1^2
    expr_original = 2.0 * s_vars[0] + 3.0 * s_vars[1]**2
    
    poly_list_intermediate = sympy2poly(expr_original, s_vars, _PSI_GLOBAL, _CLMO_GLOBAL, _ENCODE_DICT_GLOBAL)
    
    expr_reconstructed = poly2sympy(poly_list_intermediate, s_vars, _PSI_GLOBAL, _CLMO_GLOBAL)
    
    assert sp.simplify(expr_original - expr_reconstructed) == 0

    poly_list_final = sympy2poly(expr_reconstructed, s_vars, _PSI_GLOBAL, _CLMO_GLOBAL, _ENCODE_DICT_GLOBAL)
    assert _compare_poly_lists(poly_list_intermediate, poly_list_final)

    # P(x) = (1+2j)*x_0*x_1 + (3-1j)*x_2^3
    expr_original = (1+2j)*s_vars[0]*s_vars[1] + (3-1j)*s_vars[2]**3
    
    poly_list_intermediate = sympy2poly(expr_original, s_vars, _PSI_GLOBAL, _CLMO_GLOBAL, _ENCODE_DICT_GLOBAL)
    expr_reconstructed = poly2sympy(poly_list_intermediate, s_vars, _PSI_GLOBAL, _CLMO_GLOBAL)
    poly_list_final = sympy2poly(expr_reconstructed, s_vars, _PSI_GLOBAL, _CLMO_GLOBAL, _ENCODE_DICT_GLOBAL)

    assert sp.simplify(expr_original - expr_reconstructed) == 0
    assert _compare_poly_lists(poly_list_intermediate, poly_list_final)


def test_heterogeneous_polynomial_conversion():

    terms_dict = {
        (0,0,0,0,0,0): 1.5,
        (1,0,0,0,0,0): 2.0,
        (0,1,1,0,0,0): -0.5,
        (0,0,0,3,0,0): 3.0
    }
    max_deg_poly = 3
    
    custom_poly = _create_custom_poly_list(max_deg_poly, terms_dict)
    
    expected_sympy_expr = (
        sp.Float(1.5) + 
        2.0 * s_vars[0] - 
        0.5 * s_vars[1] * s_vars[2] + 
        3.0 * s_vars[3]**3
    )
    
    generated_sympy_expr = poly2sympy(custom_poly, s_vars, _PSI_GLOBAL, _CLMO_GLOBAL)
    assert sp.simplify(generated_sympy_expr - expected_sympy_expr) == 0
    
    generated_custom_poly = sympy2poly(expected_sympy_expr, s_vars, _PSI_GLOBAL, _CLMO_GLOBAL, _ENCODE_DICT_GLOBAL)
    assert _compare_poly_lists(custom_poly, generated_custom_poly)






r"""
hiten.system.body
===========

Light-weight representation of a celestial body participating in a circular restricted three 
body problem (CR3BP) or standalone dynamical simulation.

The module defines the :pyclass:`Body` class, a minimal container that stores
basic physical quantities and plotting attributes while preserving the
hierarchical relation to a central body through the :pyattr:`Body.parent`
attribute.  Instances are used across the project to compute the mass
parameter :math:`\mu` and to provide readable identifiers in logs, plots and
high-precision calculations.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional

from hiten.utils.log_config import logger


@dataclass(frozen=True)
class Body(object):
    r"""
    Celestial body container.

    Parameters
    ----------
    name : str
        Human-readable identifier, for example "Earth" or "Sun".
    mass : float
        Gravitational mass in SI units (kilograms).
    radius : float
        Mean equatorial radius in SI units (metres).
    color : str, optional
        Hexadecimal RGB string used for visualisation. Default is ``"#000000"``.
    parent : Body, optional
        Central :pyclass:`Body` that this instance orbits.  If *None*, the
        object is treated as the primary and :pyattr:`parent` is set to the
        instance itself.

    Attributes
    ----------
    name : str
        Same as the *name* parameter.
    mass : float
        Same as the *mass* parameter.
    radius : float
        Same as the *radius* parameter.
    color : str
        Colour assigned for plotting purposes.
    parent : Body
        Central body around which this instance revolves.

    Notes
    -----
    The class performs no unit or consistency checks; the responsibility of
    providing coherent values lies with the caller.

    Returns
    -------
    None

    Raises
    ------
    None

    Examples
    --------
    >>> sun = Body("Sun", 1.98847e30, 6.957e8, color="#FDB813")
    >>> earth = Body("Earth", 5.9722e24, 6.371e6, parent=sun)
    >>> print(earth)
    Earth orbiting Sun
    """
    name: str
    mass: float
    radius: float
    color: str = "#000000"
    parent: Optional[Body] = field(default=None, init=False)
    _parent_input: Optional[Body] = field(default=None, repr=False, compare=False)

    def __post_init__(self):
        # Use object.__setattr__ to bypass frozen=True protection. This is a
        # standard pattern for initializing fields in frozen dataclasses.
        parent_to_set = self._parent_input or self
        object.__setattr__(self, 'parent', parent_to_set)

        parent_name = self.parent.name if self.parent is not self else "None"
        logger.info(f"Created Body: name='{self.name}', mass={self.mass}, radius={self.radius}, color='{self.color}', parent='{parent_name}'")

    def __str__(self) -> str:
        r"""
        Return a concise human-readable description of the body.

        Returns
        -------
        str
            ``"<name> orbiting <parent>"`` when the body orbits another, or
            ``"<name> (Primary)"`` when it is the primary.
        """
        parent_desc = f"orbiting {self.parent.name}" if self.parent is not self else "(Primary)"
        return f"{self.name} {parent_desc}"

    def __repr__(self) -> str:
        r"""
        Unambiguous representation that avoids recursion.

        Notes
        -----
        The representation of a secondary body is intentionally not evaluatable
        with :pyfunc:`eval` because it only includes the parent's name to
        prevent infinite recursion. A primary body's representation, however,
        can be evaluated.

        Returns
        -------
        str
            Python expression that describes the :pyclass:`Body` instance.
        """
        # For the parent, we show its name to avoid recursion.
        # This makes the repr not perfectly eval-able for secondaries, but it's safe.
        if self.parent is self:
            parent_repr = ""
        else:
            parent_repr = f", _parent_input=Body(name='{self.parent.name}', ...)"

        return f"Body(name={self.name!r}, mass={self.mass}, radius={self.radius}, color={self.color!r}{parent_repr})"



import json
from dataclasses import asdict
from pathlib import Path
from typing import TYPE_CHECKING

import h5py
import numpy as np

from hiten.utils.io.common import _ensure_dir, _write_dataset

if TYPE_CHECKING:
    from hiten.algorithms.poincare.centermanifold.base import CenterManifoldMap
    from hiten.system.center import CenterManifold


HDF5_VERSION = "2.0"


def save_poincare_map(
    pmap: "CenterManifoldMap",
    path: str | Path,
    *,
    compression: str = "gzip",
    level: int = 4,
) -> None:
    """Serialise *pmap* (all cached sections) to *path*."""

    if not pmap._sections:
        pmap.compute()

    path = Path(path)
    _ensure_dir(path.parent)

    with h5py.File(path, "w") as f:
        f.attrs["format_version"] = HDF5_VERSION
        f.attrs["class"] = pmap.__class__.__name__
        f.attrs["energy"] = float(pmap.energy)
        f.attrs["config_json"] = json.dumps(asdict(pmap.config))

        sec_root = f.create_group("sections")

        for coord, sec in pmap._sections.items():
            g = sec_root.create_group(str(coord))
            _write_dataset(g, "points", np.asarray(sec.points), compression=compression, level=level)
            if sec.states is not None:
                _write_dataset(g, "states", np.asarray(sec.states), compression=compression, level=level)
            g.attrs["labels_json"] = json.dumps(list(sec.labels))


def load_poincare_map_inplace(
    obj: "CenterManifoldMap",
    path: str | Path,
) -> None:
    """Populate *obj* with data stored at *path*."""

    from hiten.algorithms.poincare.centermanifold.config import _CenterManifoldMapConfig
    from hiten.algorithms.poincare.core.base import _Section

    path = Path(path)
    if not path.exists():
        raise FileNotFoundError(path)

    with h5py.File(path, "r") as f:
        obj._energy = float(f.attrs["energy"])

        cfg_json = f.attrs.get("config_json", "{}")
        obj.config = _CenterManifoldMapConfig(**json.loads(cfg_json))

        obj._sections.clear()

        sec_root = f["sections"]
        for coord in sec_root.keys():
            g = sec_root[coord]
            pts = g["points"][()]
            sts = g["states"][()] if "states" in g else np.full((pts.shape[0], 4), np.nan)
            labels_json = g.attrs.get("labels_json")
            labels = tuple(json.loads(labels_json)) if labels_json else ("q2", "p2")
            obj._sections[str(coord)] = _Section(pts, sts, labels)

        obj._section = obj._sections[obj.config.section_coord]


def load_poincare_map(path: str | Path, cm: "CenterManifold"):
    """Return a new :class:`CenterManifoldMap` loaded from *path*."""
    from hiten.algorithms.poincare.centermanifold.base import CenterManifoldMap

    path = Path(path)
    if not path.exists():
        raise FileNotFoundError(path)

    with h5py.File(path, "r") as f:
        energy = float(f.attrs["energy"])

    pmap = CenterManifoldMap(cm, energy)
    load_poincare_map_inplace(pmap, path)
    return pmap
